"""Bytecode Interpreter operations for Python 3.6
"""
from __future__ import print_function, division

import inspect
import types

from xpython.byteop.byteop25 import ByteOp25
from xpython.byteop.byteop35 import ByteOp35
from xpython.pyobj import Cell, Function, make_cell

# Gone in 3.6
del ByteOp25.MAKE_CLOSURE
del ByteOp25.CALL_FUNCTION_VAR
del ByteOp25.CALL_FUNCTION_KW
del ByteOp25.CALL_FUNCTION_VAR_KW


def identity(x):
    return x


FSTRING_CONVERSION_MAP = {
    0: identity,
    1: str,
    2: repr,
    3: ascii,
}

# Code with these co_names have an implicit .0 in them
COMPREHENSION_FN_NAMES = frozenset(
    ("<setcomp>", "<dictcomp>", "<genexpr>", "<listcomp>",)
)


class ByteOp36(ByteOp35):
    def __init__(self, vm):
        super(ByteOp36, self).__init__(vm)

    def call_function_kw(self, argc):
        namedargs = {}
        namedargs_tup = self.vm.pop()
        for name in namedargs_tup:
            namedargs[name] = self.vm.pop()

        lenPos = argc - len(namedargs_tup)
        posargs = self.vm.popn(lenPos)
        func = self.vm.pop()
        self.call_function_with_args_resolved(func, posargs, namedargs)

    def format_value(self, attr, value):
        if attr & 4:
            value = self.vm.pop()
            attr_flags = attr & 3
            if attr_flags:
                conversion_fn = FSTRING_CONVERSION_MAP.get(attr_flags, identity)
            else:
                conversion_fn = identity
        else:
            conversion_fn = FSTRING_CONVERSION_MAP.get(attr, identity)

        return str(conversion_fn(value))

    ##############################################################################
    # Order of function here is the same as in:
    # https://docs.python.org/3.6/library/dis.htmls#python-bytecode-instructions
    #
    # A note about parameter names. Generally they are the same as
    # what is described above, however there are some slight changes:
    #
    # * when a parameter name is `namei` (an int), it appears as
    #   `name` (a str) below because the lookup on co_names[namei] has
    #   already been performed in parse_byte_and_args().
    ##############################################################################

    # Changed in 3.6

    def CALL_FUNCTION_KW(self, argc):
        """
        Calls a callable object with positional (if any) and keyword
        arguments.

        argc indicates the total number of positional and
        keyword arguments. The top element on the stack contains a tuple
        of keyword argument names. Below that are keyword arguments in
        the order corresponding to the tuple. Below that are positional
        arguments, with the right-most parameter on top. Below the
        arguments is a callable object to call. CALL_FUNCTION_KW pops
        all arguments and the callable object off the stack, calls the
        callable object with those arguments, and pushes the return
        value returned by the callable object.

        Changed in version 3.6: Keyword arguments are packed in a tuple
        instead of a dictionary, argc indicates the total number of
        arguments.
        """
        return self.call_function_kw(argc)

    def MAKE_FUNCTION(self, argc):
        """
        Pushes a new function object on the stack. From bottom to top,
        the consumed stack must consist of values if the argument
        carries a specified flag value

        * 0x01 a tuple of default values for positional-only and positional-or-keyword parameters in positional order
        * 0x02 a dictionary of keyword-only parameters  default values
        * 0x04 an annotation dictionary
        * 0x08 a tuple containing cells for free variables, making a closure
          the code associated with the function (at TOS1)
        * the qualified name of the function (at TOS)
        """
        name = self.vm.pop()
        code = self.vm.pop()

        slot_names = ("closure", "annotations", "kwdefaults", "closure")
        slot = {
            "defaults": tuple(),
            "kwdefaults": {},
            "annotations": {},
            "closure": tuple(),
        }
        assert argc < 17
        parameters = []
        for i in range(4):
            if argc & 1:
                parameters.insert(0, self.vm.pop())
            argc >>= 1
            if argc == 0:
                break

        for i in range(len(parameters)):
            slot[slot_names[i]] = parameters[i]

        # FIXME: DRY with code in byteop3{2,4}.py

        globs = self.vm.frame.f_globals

        # FIXME: we should test PYTHON_VERSION to check for sanity.
        if not inspect.iscode(code) and hasattr(code, "to_native"):
            code = code.to_native()

        fn_vm = Function(
            name=name,
            code=code,
            globs=globs,
            argdefs=slot["defaults"],
            closure=slot["closure"],
            vm=self.vm,
            kwdefaults=slot["kwdefaults"],
            annotations=slot["annotations"],
        )

        if argc == 0 and code.co_name in COMPREHENSION_FN_NAMES:
            fn_vm.has_dot_zero = True

        # Python 3.4 __build_class__ is more strict about what can be a
        # function type whereas in earlier version we could get away with
        # our own kind of xpython.pyobj.Function object.
        #
        # Therefore we'll try to create a native function.
        #
        # First though, we have to convert our Cells into native cells.
        # FIXME: Cells might not be equivalent to native cells. Investigate and fix if needed.
        closure = tuple(
            [
                make_cell(cell.get()) if isinstance(cell, Cell) else cell
                for cell in slot["closure"]
            ]
        )
        try:
            fn_native = types.FunctionType(
                code,
                globals=globs,
                name=name,
                argdefs=slot["defaults"],
                closure=closure,
            )
        except:
            fn_native = fn_vm
        else:
            # types.FunctionType doesn't (yet) allow these 3.x function
            # parameters, so we have to fill them in.
            fn_native.__kwdefaults__ = slot["kwdefaults"]
            fn_native.__annotations__ = slot["annotations"]

        self.vm.fn2native[fn_native] = fn_vm

        self.vm.push(fn_native)

    # New in 3.6...

    # .. and this one is gone in 3.8
    def STORE_ANNOTATION(self, name):
        """
        Stores TOS as locals()['__annotations__'][co_names[namei]] = TOS.
        """
        self.vm.frame.f_locals["__annotations__"][name] = self.vm.pop()

    def SETUP_ASYNC_WITH(self):
        """Creates a new frame object."""
        raise self.vm.PyVMError("SETUP_ASYNC_WITH not implemented yet")

    def FORMAT_VALUE(self, flags):
        """Used for implementing formatted literal strings (f-strings). Pops
        an optional fmt_spec from the stack, then a required value. flags is
        interpreted as follows:

        * (flags & 0x03) == 0x00: value is formatted as-is.
        * (flags & 0x03) == 0x01: call str() on value before formatting it.
        * (flags & 0x03) == 0x02: call repr() on value before formatting it.
        * (flags & 0x03) == 0x03: call ascii() on value before formatting it.
        * (flags & 0x04) == 0x04: pop fmt_spec from the stack and use it, else use an empty fmt_spec.

        Formatting is performed using PyObject_Format(). The result is
        pushed on the stack.
        """
        self.vm.push(self.format_value(flags, self.vm.pop()))

    def BUILD_CONST_KEY_MAP(self, count):
        """
        The version of BUILD_MAP specialized for constant keys. count
        values are consumed from the stack. The top element on the
        stack contains a tuple of keys.
        """
        keys = self.vm.pop()
        values = self.vm.popn(count)
        kvs = dict(zip(keys, values))
        self.vm.push(kvs)

    def CALL_FUNCTION_EX(self, flags):
        """
        Calls a callable object with variable set of positional and
        keyword arguments. If the lowest bit of flags is set, the top
        of the stack contains a mapping object containing additional
        keyword arguments. Below that is an iterable object containing
        positional arguments and a callable object to
        call. BUILD_MAP_UNPACK_WITH_CALL and
        BUILD_TUPLE_UNPACK_WITH_CALL can be used for merging multiple
        mapping objects and iterables containing arguments. Before the
        callable is called, the mapping object and iterable object are
        each  unpacked  and their contents passed in as keyword and
        positional arguments respectively. CALL_FUNCTION_EX pops all
        arguments and the callable object off the stack, calls the
        callable object with those arguments, and pushes the return
        value returned by the callable object.
        """
        namedargs = self.vm.pop() if flags & 1 else {}
        posargs = self.vm.pop()
        func = self.vm.pop()
        self.call_function_with_args_resolved(func, posargs, namedargs)

    def SETUP_ANNOTATIONS(self):
        """
        Checks whether __annotations__ is defined in locals(), if not it
        is set up to an empty dict. This opcode is only emitted if a
        class or module body contains variable annotations
        statically.
        """
        if "__annotations__" not in self.vm.frame.f_locals:
            self.vm.frame.f_locals["__annotations__"] = {}

    def BUILD_STRING(self, count):
        """
        The version of BUILD_MAP specialized for constant keys. count
        values are consumed from the stack. The top element on the
        stack contains a tuple of keys.
        """
        values = self.vm.popn(count)
        self.vm.push("".join(values))

    def BUILD_TUPLE_UNPACK_WITH_CALL(self, count):
        """
        This is similar to BUILD_TUPLE_UNPACK, but is used for f(*x, *y,*z)
        call syntax. The stack item at position count + 1 should be the
        corresponding callable f.
        """
        parameter_tuples = self.vm.popn(count)
        parameters = [
            parameter for sublist in parameter_tuples for parameter in sublist
        ]
        self.vm.push(parameters)



import itertools
import os
import sys

AUTO_DEBUG = os.getenv('XPYTHON_DEBUG')
RECORD_INTERPLEVEL_TRACEBACK = True

def strerror(errno):
    """Translate an error code to a unicode message string."""
    from pypy.module._codecs.locale import str_decode_locale_surrogateescape
    utf8, lgt = str_decode_locale_surrogateescape(os.strerror(errno))
    return utf8, lgt

class OperationError(Exception):
    """Interpreter-level exception that signals an exception that should be
    sent to the application level.

    OperationError instances have three attributes (and no .args),
    w_type, _w_value and _application_traceback, which contain the wrapped
    type and value describing the exception, and a chained list of
    PyTraceback objects making the application-level traceback.
    """

    _w_value = None
    _application_traceback = None
    _context_recorded = False

    def __init__(self, w_type, w_value, tb=None):
        self.setup(w_type, w_value)
        self._application_traceback = tb

    def setup(self, w_type, w_value=None):
        assert w_type is not None
        self.w_type = w_type
        self._w_value = w_value
        self.debug_excs = []

    def clear(self, space):
        # XXX remove this method.  The point is that we cannot always
        # hack at 'self' to clear w_type and _w_value, because in some
        # corner cases the OperationError will be used again: see
        # test_interpreter.py:test_with_statement_and_sys_clear.
        pass

    def match(self, space, w_check_class):
        "Check if this application-level exception matches 'w_check_class'."
        return space.exception_match(self.w_type, w_check_class)

    def _async(self, space):
        "Check if this is an exception that should better not be caught."
        return (self.match(space, space.w_SystemExit) or
                self.match(space, space.w_KeyboardInterrupt))
        # note: an extra case is added in OpErrFmtNoArgs

    def __str__(self):
        "Convenience for tracebacks."
        s = self._w_value
        space = getattr(self.w_type, 'space', None)
        if space is not None:
            if self.__class__ is not OperationError and s is None:
                s, lgt = self._compute_value(space)
            try:
                s = space.text_w(s)
            except Exception:
                pass
        return '[%s: %s]' % (self.w_type, s)

    def __repr__(self):
        "NOT_RPYTHON"
        return 'OperationError(%s)' % (self.w_type)

    def errorstr(self, space, use_repr=False):
        "The exception class and value, as a string."
        if not use_repr:    # see write_unraisable()
            self.normalize_exception(space)
        w_value = self.get_w_value(space)
        if space is None:
            # this part NOT_RPYTHON
            exc_typename = str(self.w_type)
            exc_value = str(w_value)
        else:
            exc_typename = space.text_w(
                space.getattr(self.w_type, space.newtext('__name__')))
            if space.is_w(w_value, space.w_None):
                exc_value = ""
            else:
                try:
                    if use_repr:
                        exc_value = space.text_w(space.repr(w_value))
                    else:
                        exc_value = space.text_w(space.str(w_value))
                except OperationError:
                    # oups, cannot __str__ the exception object
                    exc_value = ("<exception %s() failed>" %
                                 ("repr" if use_repr else "str"))
        if not exc_value:
            return exc_typename
        else:
            return '%s: %s' % (exc_typename, exc_value)

    def record_interpreter_traceback(self):
        """Records the current traceback inside the interpreter.
        This traceback is only useful to debug the interpreter, not the
        application."""
        if RECORD_INTERPLEVEL_TRACEBACK:
            self.debug_excs.append(sys.exc_info())

    def normalize_exception(self, space):
        """Normalize the OperationError.  In other words, fix w_type and/or
        w_value to make sure that the __class__ of w_value is exactly w_type.
        """
        #
        # This method covers all ways in which the Python statement
        # "raise X, Y" can produce a valid exception type and instance.
        #
        # In the following table, 'Class' means a subclass of BaseException
        # and 'inst' is an instance of either 'Class' or a subclass of it.
        #
        # The flow object space only deals with non-advanced case.
        #
        #  input (w_type, w_value)... becomes...                advanced case?
        # ---------------------------------------------------------------------
        #  (Class, None)              (Class, Class())                no
        #  (Class, inst)              (inst.__class__, inst)          no
        #  (Class, tuple)             (Class, Class(*tuple))          yes
        #  (Class, x)                 (Class, Class(x))               no
        #  (inst, None)               (inst.__class__, inst)          no
        #
        w_type = self.w_type
        w_value = self.get_w_value(space)

        if space.exception_is_valid_obj_as_class_w(w_type):
            # this is for all cases of the form (Class, something)
            if space.is_w(w_value, space.w_None):
                # raise Type: we assume we have to instantiate Type
                w_value = space.call_function(w_type)
                w_type = self._exception_getclass(space, w_value)
            else:
                w_valuetype = space.exception_getclass(w_value)
                if space.exception_issubclass_w(w_valuetype, w_type):
                    # raise Type, Instance: let etype be the exact type of value
                    w_type = w_valuetype
                else:
                    if space.isinstance_w(w_value, space.w_tuple):
                        # raise Type, tuple: assume the tuple contains the
                        #                    constructor args
                        w_value = space.call(w_type, w_value)
                    else:
                        # raise Type, X: assume X is the constructor argument
                        w_value = space.call_function(w_type, w_value)
                    w_type = self._exception_getclass(space, w_value)
            if self._application_traceback:
                from pypy.interpreter.pytraceback import PyTraceback
                from pypy.module.exceptions.interp_exceptions import W_BaseException
                tb = self._application_traceback
                if (isinstance(w_value, W_BaseException) and
                    isinstance(tb, PyTraceback)):
                    # traceback hasn't escaped yet
                    w_value.w_traceback = tb
                else:
                    # traceback has escaped
                    space.setattr(w_value, space.newtext("__traceback__"),
                                  self.get_w_traceback(space))
        else:
            # the only case left here is (inst, None), from a 'raise inst'.
            w_inst = w_type
            w_instclass = self._exception_getclass(space, w_inst)
            if not space.is_w(w_value, space.w_None):
                raise oefmt(space.w_TypeError,
                            "instance exception may not have a separate value")
            w_value = w_inst
            w_type = w_instclass

        self.w_type = w_type
        self._w_value = w_value

    def _exception_getclass(self, space, w_inst, what="exceptions"):
        w_type = space.exception_getclass(w_inst)
        if not space.exception_is_valid_class_w(w_type):
            raise oefmt(space.w_TypeError,
                        "%s must derive from BaseException, not %N",
                        what, w_type)
        return w_type

    def write_unraisable(self, space, where, w_object=None,
                         with_traceback=False, extra_line=''):
        # Note: since Python 3.5, unraisable exceptions are always
        # printed with a traceback.  Setting 'with_traceback=False'
        # only asks for a different format, starting with the message
        # "Exception Xxx ignored".
        if w_object is None:
            objrepr = ''
        else:
            try:
                objrepr = space.text_w(space.repr(w_object))
            except OperationError:
                objrepr = "<object repr() failed>"
        #
        try:
            try:
                self.normalize_exception(space)
            except OperationError:
                pass
            w_t = self.w_type
            w_v = self.get_w_value(space)
            w_tb = self.get_w_traceback(space)
            if where or objrepr:
                if with_traceback:
                    first_line = 'From %s%s:\n' % (where, objrepr)
                else:
                    first_line = 'Exception ignored in: %s%s\n' % (
                        where, objrepr)
            else:
                # Note that like CPython, we don't normalize the
                # exception here.  So from `'foo'.index('bar')` you get
                # "Exception ValueError: 'substring not found' in x ignored"
                # but from `raise ValueError('foo')` you get
                # "Exception ValueError: ValueError('foo',) in x ignored"
                first_line = ''
            space.appexec([space.newtext(first_line),
                           space.newtext(extra_line),
                           w_t, w_v, w_tb],
            """(first_line, extra_line, t, v, tb):
                import sys
                sys.stderr.write(first_line)
                if extra_line:
                    sys.stderr.write(extra_line)
                import traceback
                traceback.print_exception(t, v, tb)
            """)
        except OperationError:
            pass   # ignored

    def get_w_value(self, space):
        w_value = self._w_value
        if w_value is None:
            value, lgt = self._compute_value(space)
            self._w_value = w_value = space.newtext(value, lgt)
        return w_value

    def _compute_value(self, space):
        raise NotImplementedError

    def get_traceback(self):
        """Calling this marks the PyTraceback as escaped, i.e. it becomes
        accessible and inspectable by app-level Python code.  For the JIT.
        Note that this has no effect if there are already several traceback
        frames recorded, because in this case they are already marked as
        escaping by executioncontext.leave() being called with
        got_exception=True.
        """
        from pypy.interpreter.pytraceback import PyTraceback
        tb = self._application_traceback
        if tb is not None and isinstance(tb, PyTraceback):
            tb.frame.mark_as_escaped()
        return tb

    def has_any_traceback(self):
        return self._application_traceback is not None

    def set_cause(self, space, w_cause):
        if w_cause is None:
            return
        # ensure w_cause is of a valid type
        if space.is_none(w_cause):
            pass
        else:
            self._exception_getclass(space, w_cause, "exception causes")
        w_value = self.get_w_value(space)
        space.setattr(w_value, space.newtext("__cause__"), w_cause)

    def get_w_traceback(self, space):
        """Return a traceback or w_None. """
        tb = self.get_traceback()
        if tb is None:
            return space.w_None
        return tb

    def got_any_traceback(self):
        return self._application_traceback is not None

    def set_traceback(self, traceback):
        """Set the current traceback."""
        self._application_traceback = traceback

    def remove_traceback_module_frames(self, *module_names):
        from pypy.interpreter.pytraceback import PyTraceback
        tb = self._application_traceback
        while tb is not None and isinstance(tb, PyTraceback):
            if tb.frame.pycode.co_filename not in module_names:
                break
            tb = tb.next
        self._application_traceback = tb

    def record_context(self, space, ec):
        """Record a __context__ for this exception if one exists.
        """
        if self._context_recorded:
            return
        last = ec.sys_exc_info()
        try:
            if last is not None:
                self.chain_exceptions(space, last)
        finally:
            self._context_recorded = True

    def chain_exceptions(self, space, context):
        """Attach another OperationError as __context__."""
        self.normalize_exception(space)
        w_value = self.get_w_value(space)
        context.normalize_exception(space)
        w_context = context.get_w_value(space)
        if not space.is_w(w_value, w_context):
            _break_context_cycle(space, w_value, w_context)
            space.setattr(w_value, space.newtext('__context__'), w_context)

    def chain_exceptions_from_cause(self, space, exception):
        # XXX does this code really make sense?
        self.chain_exceptions(space, exception)
        self.set_cause(space, exception.get_w_value(space))
        self.record_context(space, space.getexecutioncontext())

    # A simplified version of _PyErr_TrySetFromCause, which returns a
    # new exception of the same class, but with another error message.
    # This only works for exceptions which have just a single message,
    # and no other attribute.
    # Otherwise the same OperationError is returned.
    def try_set_from_cause(self, space, message):
        from pypy.module.exceptions.interp_exceptions import W_BaseException
        self.normalize_exception(space)
        w_value = self.get_w_value(space)
        if not isinstance(w_value, W_BaseException):
            return self
        exc = w_value
        # "args" should be empty or contain a single string
        if len(exc.args_w) == 0:
            pass
        elif len(exc.args_w) == 1:
            if not space.isinstance_w(exc.args_w[0], space.w_unicode):
                return self
        else:
            return self
        # No instance attribute.
        if exc.w_dict and space.is_true(exc.w_dict):
            return self
        # Try to create the new exception.
        try:
            new_error = oefmt(space.type(w_value),
                              "%s (%T: %S)", message, w_value, w_value)
            new_error.normalize_exception(space)
            new_error.set_cause(space, w_value)
            # Copy the traceback, but it does not escape.
            new_error.set_traceback(self._application_traceback)
        except OperationError:
            # Return the original error
            return self
        return new_error


def _break_context_cycle(space, w_value, w_context):
    """Break reference cycles in the __context__ chain.

    This is O(chain length) but context chains are usually very short
    """
    while True:
        w_next = space.getattr(w_context, space.newtext('__context__'))
        if space.is_w(w_next, space.w_None):
            break
        if space.is_w(w_next, w_value):
            space.setattr(w_context, space.newtext('__context__'), space.w_None)
            break
        w_context = w_next


class ClearedOpErr:
    def __init__(self, space):
        self.operr = OperationError(space.w_None, space.w_None)

def get_cleared_operation_error(space):
    return space.fromcache(ClearedOpErr).operr

# ____________________________________________________________
# optimization only: avoid the slowest operation -- the string
# formatting with '%' -- in the common case were we don't
# actually need the message.  Only supports %s and %d.

_fmtcache = {}
_fmtcache2 = {}
_FMTS = tuple('8NRSTds')

def decompose_valuefmt(valuefmt):
    """Returns a tuple of string parts extracted from valuefmt,
    and a tuple of format characters."""
    formats = []
    parts = valuefmt.split('%')
    i = 1
    while i < len(parts):
        if parts[i].startswith(_FMTS):
            formats.append(parts[i][0])
            parts[i] = parts[i][1:]
            i += 1
        elif parts[i] == '':    # support for '%%'
            parts[i-1] += '%' + parts[i+1]
            del parts[i:i+2]
        else:
            fmts = '%%%s or %%%s' % (', %'.join(_FMTS[:-1]), _FMTS[-1])
            raise ValueError("invalid format string (only %s supported)" %
                             fmts)
    assert len(formats) > 0, "unsupported: no % command found"
    return tuple(parts), tuple(formats)

def get_operrcls2(valuefmt):
    strings, formats = decompose_valuefmt(valuefmt)
    assert len(strings) == len(formats) + 1
    try:
        OpErrFmt = _fmtcache2[formats]
    except KeyError:
        from rpython.rlib.unroll import unrolling_iterable
        attrs = ['x%d' % i for i in range(len(formats))]
        entries = unrolling_iterable(zip(itertools.count(), formats, attrs))

        class OpErrFmt(OperationError):
            def __init__(self, w_type, strings, *args):
                assert len(args) == len(strings) - 1
                self.xstrings = strings
                for i, _, attr in entries:
                    setattr(self, attr, args[i])
                self.setup(w_type)

            def _compute_value(self, space):
                lst = [None] * (len(formats) + len(formats) + 1)
                lgt = 0
                for i, fmt, attr in entries:
                    lst[i + i] = self.xstrings[i]
                    lgt += len(self.xstrings[i])
                    value = getattr(self, attr)
                    if fmt == 'd':
                        result = str(value)
                        lgt += len(result)
                    elif fmt == 'R':
                        s = space.repr(value)
                        result = space.utf8_w(s)
                        lgt += space.len_w(s)
                    elif fmt == 'S':
                        s = space.str(value)
                        result = space.utf8_w(s)
                        lgt += space.len_w(s)
                    # FIXME rocky:  reinstate when needed
                    # elif fmt == 'T':
                    #     result = space.type(value).name
                    #     lgt += rutf8.codepoints_in_utf8(result)
                    elif fmt == 'N':
                        result = value.getname(space)
                        lgt += len(result)
                    elif fmt == '8':
                        # u'str\uxxxx' -> 'str\xXX\xXX' -> u"'str\xXX\xXX'"
                        from pypy.interpreter import unicodehelper
                        result, _lgt, pos  = unicodehelper.str_decode_utf8(
                            value, 'replace', True,
                            unicodehelper.decode_never_raise, True)
                        lgt += _lgt
                    elif isinstance(value, unicode):
                        # 's'
                        result = str(value.encode('utf-8'))
                        lgt += len(value)
                    # FIXME rocky: reinstate when needed
                    # else:
                    #     result = str(value)
                    #     try:
                    #         lgt += rutf8.check_utf8(result, True)
                    #     except rutf8.CheckError as e:
                    #         lgt -= e.pos
                    lst[i + i + 1] = result
                lst[-1] = self.xstrings[-1]
                lgt += len(self.xstrings[-1])
                retval = ''.join(lst)
                return retval, lgt

        _fmtcache2[formats] = OpErrFmt
    return OpErrFmt, strings

class OpErrFmtNoArgs(OperationError):
    def __init__(self, w_type, value):
        self._value = value
        self.setup(w_type)

    def _compute_value(self, space):
        return self._value, len(self._value)

    def _async(self, space):
        # also matches a RuntimeError("maximum rec.") if the stack is
        # still almost full, because in this case it might be a better
        # idea to propagate the exception than eat it
        if (self.w_type is space.w_RecursionError and
            self._value == "maximum recursion depth exceeded"):
            return True
        return OperationError._async(self, space)

def get_operr_class(valuefmt):
    try:
        result = _fmtcache[valuefmt]
    except KeyError:
        result = _fmtcache[valuefmt] = get_operrcls2(valuefmt)
    return result

def oefmt(w_type, valuefmt, *args):
    """Equivalent to OperationError(w_type, space.newtext(valuefmt % args)).
    More efficient in the (common) case where the value is not actually
    needed. Note that in the py3k branch the exception message will
    always be unicode.

    Supports the standard %s and %d formats, plus the following:

    %8 - The result of arg.decode('utf-8')
    %N - The result of w_arg.getname(space)
    %R - The result of space.utf8_w(space.repr(w_arg))
    %S - The result of space.utf8_w(space.str(w_arg))
    %T - The result of space.type(w_arg).name

    """
    if not len(args):
        return OpErrFmtNoArgs(w_type, valuefmt)
    OpErrFmt, strings = get_operr_class(valuefmt)
    return OpErrFmt(w_type, strings, *args)

def exception_from_errno(space, w_type, errno):
    msg, lgt = strerror(errno)
    w_error = space.call_function(w_type, space.newint(errno),
                                  space.newtext(msg, lgt))
    return OperationError(w_type, w_error)

def exception_from_saved_errno(space, w_type):
    from rpython.rlib.rposix import get_saved_errno
    errno = get_saved_errno()
    return exception_from_errno(space, w_type, errno)

def new_exception_class(space, name, w_bases=None, w_dict=None):
    """Create a new exception type.
    @param name: the name of the type.
    @param w_bases: Either an exception type, or a wrapped tuple of
                    exception types.  default is space.w_Exception.
    @param w_dict: an optional dictionary to populate the class __dict__.
    """
    if '.' in name:
        module, name = name.rsplit('.', 1)
    else:
        module = None
    if w_bases is None:
        w_bases = space.newtuple([space.w_Exception])
    elif not space.isinstance_w(w_bases, space.w_tuple):
        w_bases = space.newtuple([w_bases])
    if w_dict is None:
        w_dict = space.newdict()
    w_exc = space.call_function(
        space.w_type, space.newtext(name), w_bases, w_dict)
    if module:
        space.setattr(w_exc, space.newtext("__module__"), space.newtext(module))
    return w_exc

def new_import_error(space, w_msg, w_name, w_path):
    """Create a new instance of ImportError.

    The result corresponds to ImportError(msg, name=name, path=path)
    """
    return space.appexec(
        [w_msg, w_name, w_path], """(msg, name, path):
            return ImportError(msg, name=name, path=path)""")

def raise_import_error(space, w_msg, w_name, w_path):
    w_exc = new_import_error(space, w_msg, w_name, w_path)
    raise OperationError(space.w_ImportError, w_exc)



"""A main program for xpython."""

import click
import logging
import sys

from xpython import execfile
from xpython.vm import PyVMRuntimeError
from xpython.version import VERSION
from xdis import PYTHON_VERSION, IS_PYPY

def version_message():
    platform = "PyPy " if IS_PYPY else "C"
    mess = "%s running from %sPython %s" % (
        VERSION, platform, PYTHON_VERSION
    )
    return mess

@click.command()
@click.version_option(version_message(), "-V", "--version")
@click.option("-m", "--module", default=False,
              help="PATH is a module name, not a Python main program")
@click.option("-v", "--verbose", count=True,
              help="verbosity level in tracing.\n"
              "Can be supplied multiple times to increase verbosity.")
@click.option("-c", "--command-to-run",
              help="program passed in as a string", required=False)
@click.argument("path", nargs=1, type=click.Path(readable=True), required=False)
@click.argument("args", nargs=-1)
def main(module, verbose, command_to_run, path, args):
    """
    Runs Python programs or bytecode using a bytecode interpreter written in Python.
    """
    if module:
        run_fn = execfile.run_python_module
    else:
        run_fn = execfile.run_python_file

    if verbose > 1:
        level = logging.DEBUG
    elif verbose == 1:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)

    if command_to_run:
        if path or args:
            print("You must pass either a file name or a command string, not both.")
            sys.exit(4)
        path = command_to_run
        run_fn = execfile.run_python_string
    elif not path:
        print("You must pass either a file name or a command string, neither found.")
        sys.exit(4)

    try:
        run_fn(path, args)
    except PyVMRuntimeError:
        # Tracebacks and error messages should been previously printed
        sys.exit(10)
    except execfile.CannotCompileError as e:
        print(e)
        sys.exit(1)
        pass
    except execfile.NoSourceError as e:
        if verbose > 1:
            raise
        print(e)
        sys.exit(2)
        pass
    except execfile.WrongBytecodeError as e:
        if verbose > 1:
            raise
        print(e)
        sys.exit(3)
        pass
    except SystemExit:
        # Program ran sys.exit();
        # Respect that.
        raise

if __name__ == "__main__":
    main(auto_envvar_prefix="XPYTHON")



"""This program is self-checking!"""

def f1(a):
    b = 2*a
    def f2(c):
        d = 2*c
        def f3(e):
            f = 2*e
            def f4(g):
                h = 2*g
                return a+b+c+d+e+f+g+h
            return f4
        return f3
    return f2

answer = f1(3)(4)(5)(6)
assert answer == 54

# From test_audiop.py
# This is an example where we to create a function with a non-zero
# closure tuple of cells. In some versions we'll need to convert
# our Cell type into a native cell type if we want to create a native
# function.

import sys
if sys.version_info[0:1] >= (2, 7):
    def pack(width, data):
        return b''.join(v.to_bytes(width, "little", signed=True) for v in data)

    packs = {w: (lambda data, width=w: pack(width, data)) for w in (1, 2, 3, 4)}
    assert packs[3](data=[0, -1, 0x123456]) == b'\x00\x00\x00\xff\xff\xffV4\x12'



"""Bytecode Interpreter operations for PyPy 3.2
"""
from __future__ import print_function, division

from xpython.byteop.byteop32 import ByteOp32
from xpython.byteop.byteoppypy import ByteOpPyPy

class ByteOp32PyPy(ByteOp32, ByteOpPyPy):
    def __init__(self, vm):
        super(ByteOp32PyPy, self).__init__(vm)



class NullContext(object):
    def __enter__(self):
        l.append('i')
        # __enter__ usually returns self, but doesn't have to.
        return 17

    def __exit__(self, exc_type, exc_val, exc_tb):
        l.append('o')
        return False

l = []
for i in range(3):
    with NullContext() as val:
        assert val == 17
        l.append('w')
    l.append('e')
l.append('r')
s = ''.join(l)
assert s == "iwoeiwoeiwoer"



"""This program is self-checking!"""

def replace_globals(f, new_globals):
    import sys

    if sys.version_info.major == 2:
        args = [
            f.func_code,
            new_globals,
            f.func_name,
            f.func_defaults,
            f.func_closure,
        ]
    else:
        args = [
            f.__code__,
            new_globals,
            f.__name__,
            f.__defaults__,
            f.__closure__,
        ]
    if hasattr(f, "_vm"):
        name = args.remove(args[2])
        args.insert(0, name)
        args.append(f._vm)
    return type(lambda: None)(*args)


def f():
    assert g() == 2
    assert a == 1


def g():
    return a  # a is in the builtins and set to 2


# g and f have different builtins that both provide ``a``.
g = replace_globals(g, {"__builtins__": {"a": 2}})
f = replace_globals(f, {"__builtins__": {"a": 1}, "g": g})


f()



# This test is good for checking
# WITH_CLEANUP (< 3.5) or WITH_CLEANUP_START/WITH_CLEANUP_FINISH
# In particular there is a custom __exit__ handler

class NullContext(object):
    def __enter__(self):
        l.append('i')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        l.append('o')
        return False

l = []
for i in range(3):
    with NullContext():
        l.append('w')
        if i % 2:
           continue
        l.append('z')
    l.append('e')

l.append('r')
s = ''.join(l)
assert s == "iwzoeiwoiwzoer"



"""Bytecode Interpreter operations for PyPy in general (all versions)

Specific PyPy versions i.e. PyPy 2.7, 3.2, 3.5, and 3.6 inherit this.
"""
from __future__ import print_function, division


class ByteOpPyPy(object):
    def LOOKUP_METHOD(self, name):
        """
        For now, we'll assume this is the same as LOAD_ATTR:

        Replaces TOS with getattr(TOS, co_names[namei]).
        Note: name = co_names[namei] set in parse_byte_and_args()

        """
        obj = self.vm.pop()
        val = getattr(obj, name)
        self.vm.push(val)

    def CALL_METHOD(self, argc):
        """
        For now, we'll assume this is like CALL_FUNCTION:

        Calls a callable object.
        The low byte of argc indicates the number of positional
        arguments, the high byte the number of keyword arguments.
        ...
        """
        return self.call_function(argc, [], {})



# This tests the use of the "global" statement.

# This program has the simplest of subroutine calls
# so generally it is a good program to try
# when expanding the interpreter such as for a new
# Python version.

"""This program is self-checking!"""

global xyz
xyz=2106

def abc():
    global xyz
    xyz+=1
    assert xyz == 2107, "Midst failed"

assert xyz == 2106, "Pre failed"
abc()
assert xyz == 2107, "Post failed"



# Test various kinds of comprehensions.
# In various Pythons there is this .0 parameter
# which isn't listed in the signature by inpsect.
"""This program is self-checking!"""

x = [z*z for z in range(5)]
assert x == [0, 1, 4, 9, 16]

x = {z:z*z for z in range(5)}
assert x == {0:0, 1:1, 2:4, 3:9, 4:16}

x = {z*z for z in range(5)}
assert x == {0, 1, 4, 9, 16}



"""Testing intepreting a different bytcode from the one the Python interpreter is using."""
from __future__ import print_function

try:
    import vmtest
except ImportError:
    from . import vmtest

from glob import iglob
import os
from xdis import PYTHON_VERSION

# In the dictionary below,a
# The key is a Python version and
# the value is a list of Python bytecode
# different from itself that it should be able to handle
COMPATABILITY_MATRIX = {
    2.7: (2.6, 2.5),
    3.2: (),
    3.3: (3.2,),
    # More or less punt on these for now
    3.4:  (),
    3.5: (),
    3.6: (),
    3.7: (),
}

# In addition, PyPy and CPython *should* be inter compatible,
# since we should take pains to handle PyPy's LOOKUP_METHOD and
# CALL_METHOD.

# How to increase compatablity above.
# 1. Write a replacement for builtin build_class
# 2. The `inspect` module is used to grock function calls.
#    it has checks for iscode and so on. Change those to
#    be use duck typing.

class TestCompat(vmtest.VmTestCase):

    if not os.environ.get("SKIP_COMPAT", False):
        def test_cross_version_compatablity(self):
            compatible_versions = COMPATABILITY_MATRIX.get(PYTHON_VERSION, ())
            # import logging
            # logging.basicConfig(level=logging.DEBUG)
            for version in compatible_versions:
                os.chdir(os.path.join(vmtest.srcdir, "bytecode-%s" % version))
                for bytecode_file in iglob("*.pyc"):
                    # print(bytecode_file, version)
                    self.assert_runs_ok(bytecode_file, arg_type="bytecode-file")

if __name__ == "__main__":
    import unittest
    unittest.main()

    # t = TestCompat("test_cross_version_compatablity")
    # t.test_cross_version_compatablity()



from contextlib import contextmanager

def inner():
    yield "I'm inner!"

def foo():
    yield from inner()

    @contextmanager
    def cmgr():
        yield "Context Manager!"
    raise StopIteration(cmgr())

def main():
    with (yield from foo()) as x:
        print(x)

def run(fn, *args):
    x = fn(*args)
    while True:
        try:
            print(next(x))
        except StopIteration as e:
            return e.value
run(main)



# Tests a bug where a function doesn't get
# builtin-functions

"""This program is self-checking!"""

def replace_globals(f, new_globals):
    import sys


    if sys.version_info.major == 2:
        args = [
            f.func_code,
            new_globals,
            f.func_name,
            f.func_defaults,
            f.func_closure,
        ]
    else:
        args = [
            f.__code__,
            new_globals,
            f.__name__,
            f.__defaults__,
            f.__closure__,
        ]
    if hasattr(f, '_vm'):
        name = args.remove(args[2])
        args.insert(0, name)
        args.append(f._vm)
    return type(lambda: None)(*args)


def f(NameError=NameError, AssertionError=AssertionError):
    # capture NameError and AssertionError early because
    #  we are deleting the builtins
    None
    try:
        sum
    except NameError:
        pass
    else:
        raise AssertionError('sum in the builtins')


f = replace_globals(f, {})  # no builtins provided
f()



"""A variant of VirtualMachine that adds a callback.
This can be used in a debugger or profiler.
"""

import six
from xdis import PYTHON_VERSION, IS_PYPY
from xpython.vm import (
    byteint,
    PyVM,
    PyVMError,
)
from xpython.pyobj import traceback_from_frame

import logging

log = logging.getLogger(__name__)

PyVMEVENT_INSTRUCTION = 1
PyVMEVENT_LINE = 2
PyVMEVENT_CALL = 4
PyVMEVENT_RETURN = 8
PyVMEVENT_EXCEPTION = 16
PyVMEVENT_YIELD = 32


class PyVMTraced(PyVM):
    def __init__(
        self,
        callback,
        python_version=PYTHON_VERSION,
        is_pypy=IS_PYPY,
        vmtest_testing=False,
        event_flags=31,
    ):
        super().__init__(python_version, is_pypy, vmtest_testing)
        self.callback = callback
        self.event_flags = event_flags

        # TODO:
        # * add events of interest
        # * add offset breakpoints
        # * add line number breakopints
        # * add function breakpoints

    def run_frame(self, frame):
        """Run a frame until it returns (somehow).

        Exceptions are raised, the return value is returned.

        """
        if frame.f_lasti == -1:
            # We were started new, not yielded back from
            frame.f_lasti = 0
            frame.fallthrough = (
                False  # Don't increment before fetching next instruction
            )
            byteCode = None
            last_i = frame.f_back.f_lasti if frame.f_back else -1
            self.push_frame(frame)
            if self.event_flags & PyVMEVENT_CALL:
                self.callback("call", last_i, "CALL", byteCode, frame.f_lineno, [], self)
                pass
        else:
            byteCode = byteint(frame.f_code.co_code[frame.f_lasti])
            self.push_frame(frame)
            if self.event_flags & PyVMEVENT_YIELD:
                self.callback("yield", opoffset, byteName, "YIELD_VALUE", frame.f_lineno, [], self)
                pass
            # byteCode == opcode["YIELD_VALUE"]?

        self.linestarts = dict(self.opc.findlinestarts(frame.f_code, dup_lines=True))

        opoffset = 0
        while True:

            (
                byteName,
                byteCode,
                intArg,
                arguments,
                opoffset,
                line_number,
            ) = self.parse_byte_and_args(byteCode)

            if log.isEnabledFor(logging.INFO):
                self.log(byteName, intArg, arguments, opoffset, line_number)

            if line_number is not None and self.event_flags & (
                PyVMEVENT_LINE | PyVMEVENT_INSTRUCTION
            ):
                self.callback("line", opoffset, byteName, byteCode, line_number, arguments, self)
            elif self.event_flags & PyVMEVENT_INSTRUCTION:
                self.callback("instruction", opoffset, byteName, byteCode, line_number, arguments, self)

            # When unwinding the block stack, we need to keep track of why we
            # are doing it.
            why = self.dispatch(byteName, intArg, arguments, opoffset, line_number)
            if why == "exception":
                # Deal with exceptions encountered while executing the op.
                if not self.in_exception_processing:
                    self.last_traceback = traceback_from_frame(self.frame)
                    self.in_exception_processing = True

            if why == "reraise":
                why = "exception"

            if why != "yield":
                while why and frame.block_stack:
                    # Deal with any block management we need to do.
                    why = self.manage_block_stack(why)

            if why:
                break

        # TODO: handle generator exception state

        if why == "exception":
            if self.event_flags & PyVMEVENT_EXCEPTION:
                self.callback(
                    "exception", opoffset, byteName, byteCode, line_number, self.last_exception, self
                )
            elif self.event_flags & PyVMEVENT_RETURN:
                self.callback("return", opoffset, byteName, byteCode, line_number, self.return_value, self)
            pass

        self.pop_frame()

        if why == "exception":
            if self.last_exception and self.last_exception[0]:
                six.reraise(*self.last_exception)
            else:
                raise PyVMError("Borked exception recording")
            # if self.exception and .... ?
            # log.error("Haven't finished traceback handling, nulling traceback information for now")
            # six.reraise(self.last_exception[0], None)

        self.in_exception_processing = False
        if self.event_flags & PyVMEVENT_RETURN:
            self.callback("return", opoffset, byteName, byteCode, line_number, self.return_value, self)

        return self.return_value



"""Bytecode Interpreter operations for Python 3.5
"""
from __future__ import print_function, division

import types
from xpython.byteop.byteop25 import ByteOp25
from xpython.byteop.byteop32 import ByteOp32
from xpython.byteop.byteop34 import ByteOp34

# Gone in 3.5
del ByteOp25.STORE_MAP
del ByteOp32.WITH_CLEANUP


class ByteOp35(ByteOp34):
    def __init__(self, vm):
        super(ByteOp35, self).__init__(vm)

    def build_container_flat(self, count, container_fn):
        elts = self.vm.popn(count)
        self.vm.push(container_fn(e for l in elts for e in l))

    # Changed in 3.5

    def BUILD_MAP(self, count):
        """
        Pushes a new dictionary object onto the stack. Pops 2 * count
        items so that the dictionary holds count entries: {..., TOS3:
        TOS2, TOS1: TOS}.

        Changed in version 3.5: The dictionary is created from stack
        items instead of creating an empty dictionary pre-sized to
        hold count items.
        """
        kvs = self.vm.popn(count * 2)
        self.vm.push(dict(kvs[i : i + 2] for i in range(0, len(kvs), 2)))

    # New in 3.5

    def BUILD_TUPLE_UNPACK(self, count):
        """
        Pops count iterables from the stack, joins them in a single tuple,
        and pushes the result. Implements iterable unpacking in
        tuple displays (*x, *y, *z).
        """
        self.build_container_flat(count, tuple)

    def BUILD_LIST_UNPACK(self, count):
        """
        This is similar to BUILD_TUPLE_UNPACK, but pushes a list instead of tuple.
        Implements iterable unpacking in list displays [*x, *y, *z].
        """
        self.build_container_flat(count, list)

    def BUILD_SET_UNPACK(self, count):
        """
        Pops count mappings from the stack, merges them into a single
        dictionary, and pushes the result. Implements dictionary
        unpacking in dictionary displays {**x, **y, **z}.
        """
        self.build_container_flat(count, set)

    def BUILD_MAP_UNPACK(self, count):
        """
        Pops count iterables from the stack, joins them in a single tuple,
        and pushes the result. Implements iterable unpacking in
        tuple displays (*x, *y, *z).
        """
        # Note: this isn't the same thing as build_container_flat
        elts = self.vm.popn(count)
        result = {}
        for d in elts:
            result.update(d)
        self.vm.push(result)

    def BUILD_MAP_UNPACK_WITH_CALL(self, oparg):
        """
        This is similar to BUILD_MAP_UNPACK, but is used for f(**x, **y,
        **z) call syntax. The lowest byte of oparg is the count of
        mappings, the relative position of the corresponding callable
        f is encoded in the second byte of oparg.
        """
        fn_pos, count = divmod(oparg, 256)
        elts = self.vm.popn(count)
        if elts:
            kwargs = {k: v for m in elts for k, v in m.items()}
        else:
            kwargs = None
        posargs = self.vm.pop()
        func = self.vm.pop(fn_pos)

        # Put everything in the right order for CALL_FUNCTION_EX
        self.vm.push(func)
        self.vm.push(posargs)
        if kwargs:
            self.vm.push(kwargs)

    # Coroutine opcodes

    def GET_AWAITABLE(self):
        """
        Implements TOS = get_awaitable(TOS), where get_awaitable(o)
        returns o if o is a coroutine object or a generator object
        with the CO_ITERABLE_COROUTINE flag, or resolves
        o.__await__.
        """
        # # Adapted from PyPy 3.6 v. 7.3.1
        # from xpython.interpreter.generator import get_awaitable_iter
        # from xpython.interpreter.generator import Coroutine
        # iterable = self.vm.pop()
        # iter = get_awaitable_iter(self.space, iterable)
        # if isinstance(iter, Coroutine):
        #     if iter.get_delegate() is not None:
        #         # 'w_iter' is a coroutine object that is being awaited,
        #         # '.w_yielded_from' is the current awaitable being awaited on.
        #         raise RuntimeError("coroutine is being awaited already")

        raise self.vm.PyVMError("GET_AWAITABLE not implemented yet")

    def GET_AITER(self):
        """
        Implements TOS = get_awaitable(TOS.__aiter__()). See GET_AWAITABLE
        for details about get_awaitable
        """
        raise self.vm.PyVMError("GET_AITER not implemented yet")
        # TOS = self.vm.pop()
        # self.vm.push(get_awaitable(TOS.__aiter__()))

    def GET_ANEXT(self):
        """
        Implements PUSH(get_awaitable(TOS.__anext__())). See GET_AWAITABLE
        for details about get_awaitable
        """
        raise self.vm.PyVMError("GET_ANEXT not implemented yet")
        # TOS = self.vm.pop()
        # self.vm.push(get_awaitable(TOS.__anext()))

    def BEFORE_ASYNC_WITH(self):
        raise self.vm.PyVMError("BEFORE_ASYNC_WITH not implemented yet")
        return

    def GET_YIELD_FROM_ITER(self):
        """
        If TOS is a generator iterator or coroutine object it is left as
        is. Otherwise, implements TOS = iter(TOS).
        """
        TOS = self.vm.top()
        if isinstance(TOS, types.GeneratorType) or isinstance(TOS, types.CoroutineType):
            return
        TOS = self.vm.pop()
        self.vm.push(iter(TOS))

    def WITH_CLEANUP_START(self):
        """Cleans up the stack when a with statement block exits.

        TOS is the context manager s __exit__() bound method. Below
        TOS are 1 - 3 values indicating how/why the finally clause was
        entered:

        * SECOND = None
        * (SECOND, THIRD) = (WHY_{RETURN,CONTINUE}), retval
        * SECOND = WHY_*; no retval below it
        * (SECOND, THIRD, FOURTH) = exc_info()

        In the last case, TOS(SECOND, THIRD, FOURTH) is called,
        otherwise TOS(None, None, None). Pushes SECOND and result of the call
        to the stack.  Cleans up the stack when a `with` statement block
        exits. TOS is the context manager's `__exit__()` bound method.
        """
        second = third = fourth = None
        TOS = self.vm.top()
        if TOS is None:
            exit_method = self.vm.pop(1)
        elif isinstance(TOS, str):
            # FIXME: This code does something funky with pushing "continue"
            # See the comment under CONTINUE_LOOP.
            # jump addresses on the frame stack. As a result, we need to
            # set up here something to make END_FINALLY work and remove
            # the jump address. This means that the comment or semantics
            # described above isn't strictly correct.
            if TOS in ("return", "continue"):
                exit_method = self.vm.pop(2)
            else:
                exit_method = self.vm.pop(1)
        elif issubclass(TOS, BaseException):
            fourth, third, second = self.vm.popn(3)
            tp, exc, tb = self.vm.popn(3)
            exit_method = self.vm.pop()
            self.vm.push(None)
            self.vm.push(fourth, third, second)
            block = self.vm.pop_block()
            assert block.type == "except-handler"
            self.vm.push_block(block.type, block.handler, block.level - 1)
        exit_ret = exit_method(second, third, fourth)
        self.vm.push(second)
        self.vm.push(exit_ret)

    def WITH_CLEANUP_FINISH(self):
        """Pops exception type and result of "exit" function call from the stack.

        If the stack represents an exception, and the function call
        returns a  true  value, this information is "zapped" and
        replaced with a single WHY_SILENCED to prevent END_FINALLY
        from re-raising the exception. (But non-local gotos will still
        be resumed.)
        """
        exit_ret = self.vm.pop(1)
        u = self.vm.pop()
        if type(u) is type and issubclass(u, BaseException) and exit_ret:
            self.vm.push("silenced")



# FIXME: There is probably way to use importer to simplify this.  The
# problem is that historically this has been a big mess, which changes
# a lot and we are interested in supporting (some) historical versions
# of Python.

def get_byteop(vm, python_version, is_pypy):
    """Get Python byteop for given integer Python version, e.g. 2.7, 3.2, 3.5..., and the
    platform is_pypy. vm.VMError will be raised if we can't find a suitable version.
    """

    int_vers = int(python_version * 10)

    if int_vers < 30:
        if int_vers == 27:
            if is_pypy:
                from xpython.byteop.byteop27pypy import ByteOp27PyPy

                byteop = ByteOp27PyPy(vm)
            else:
                from xpython.byteop.byteop27 import ByteOp27

                byteop = ByteOp27(vm)
        elif int_vers == 26:
            from xpython.byteop.byteop26 import ByteOp26

            byteop = ByteOp26(vm)
            pass
        elif int_vers == 25:
            from xpython.byteop.byteop25 import ByteOp25

            byteop = ByteOp25(vm)
            pass
        pass
    else:
        # 3.0 or greater
        if int_vers < 35:
            if int_vers == 32:
                if is_pypy:
                    from xpython.byteop.byteop32pypy import ByteOp32PyPy

                    byteop = ByteOp32PyPy(vm)
                else:
                    from xpython.byteop.byteop32 import ByteOp32

                    byteop = ByteOp32(vm)
            elif int_vers == 33:
                from xpython.byteop.byteop33 import ByteOp33

                byteop = ByteOp33(vm)
            elif int_vers == 34:
                from xpython.byteop.byteop34 import ByteOp34

                byteop = ByteOp34(vm)
            else:
                raise vm.VMEerror("Version %s not supported" % python_version)
        else:
            if int_vers == 35:
                if is_pypy:
                    from xpython.byteop.byteop35pypy import ByteOp35PyPy

                    byteop = ByteOp35PyPy(vm)
                else:
                    from xpython.byteop.byteop35 import ByteOp35

                    byteop = ByteOp35(vm)
            elif int_vers == 36:
                if is_pypy:
                    from xpython.byteop.byteop36pypy import ByteOp36PyPy

                    byteop = ByteOp36PyPy(vm)
                else:
                    from xpython.byteop.byteop36 import ByteOp36

                    byteop = ByteOp36(vm)
            elif int_vers == 37:
                from xpython.byteop.byteop37 import ByteOp37

                byteop = ByteOp37(vm)
            else:
                raise vm.VMEerror("Version %s not supported" % python_version)
            pass
        pass
    return byteop



# This code was originally written by Darius Bacon,
# but follows code from PEP 3115 listed below.
# Rocky Bernstein did the xdis adaptions and
# added a couple of bug fixes.

from xpython.pyobj import Function, Cell
from xdis import codeType2Portable, PYTHON_VERSION, IS_PYPY


def build_class(opc, func, name, *bases, **kwds):
    """
    Like built-in __build_class__() in bltinmodule.c, but running in the
    byterun VM.

    See also: PEP 3115: https://www.python.org/dev/peps/pep-3115/ and
    https://mail.python.org/pipermail/python-3000/2007-March/006338.html
    """

    # Parameter checking...
    if not isinstance(func, Function):
        raise TypeError("func must be a function")
    if not isinstance(name, str):
        raise TypeError("name is not a string")

    metaclass = kwds.pop("metaclass", None)

    if metaclass is None:
        metaclass = type(bases[0]) if bases else type
    if isinstance(metaclass, type):
        metaclass = calculate_metaclass(metaclass, bases)

    if hasattr(metaclass, "__prepare__"):
        prepare = metaclass.__prepare__
        namespace = prepare(name, bases, **kwds)
    else:
        namespace = {}

    python_implementation = "PyPy" if IS_PYPY else "CPython"

    if not (
        opc.version == PYTHON_VERSION
        and python_implementation == opc.python_implementation
    ):
        # convert code to xdis's portable code type.
        class_body_code = codeType2Portable(func.func_code)
    else:
        class_body_code = func.func_code

    # Execute the body of func. This is the step that would go wrong if
    # we tried to use the built-in __build_class__, because __build_class__
    # does not call func, it magically executes its body directly, as we
    # do here (except we invoke our PyVM instead of CPython's).
    #
    # This behavior when interpreting bytecode that isn't the same as
    # the bytecode using in the running Python can cause a SEGV, specifically
    # between Python 3.5 running 3.4 or earlier.
    frame = func._vm.make_frame(
        code=class_body_code, f_globals=func.func_globals, f_locals=namespace
    )

    # rocky: cell is the return value of a function where?
    cell = func._vm.run_frame(frame)

    # Add any class variables that may have been added in running class_body_code.
    # See test_attribute_access.py for a simple example that needs the update below.
    namespace.update(frame.f_locals)

    cls = metaclass(name, bases, namespace)
    if isinstance(cell, Cell):
        cell.set(cls)
    return cls


def calculate_metaclass(metaclass, bases):
    "Determine the most derived metatype."
    winner = metaclass
    for base in bases:
        t = type(base)
        if issubclass(t, winner):
            winner = t
        elif not issubclass(winner, t):
            raise TypeError("metaclass conflict", winner, t)
    return winner



# -*- coding: utf-8 -*-
"""Byte Interpreter operations for Python 3.3
"""
from __future__ import print_function, division

from xpython.byteop.byteop32 import ByteOp32
from xpython.pyobj import Generator


class ByteOp33(ByteOp32):
    def __init__(self, vm):
        super(ByteOp33, self).__init__(vm)

    # Right now 3.3 is largely the same as 3.2 here. How nice!

    def YIELD_FROM(self):
        """
        Pops TOS and delegates to it as a subiterator from a generator.
        """
        u = self.vm.pop()
        x = self.vm.top()

        try:
            if not isinstance(x, Generator) or u is None:
                # Call next on iterators.
                retval = next(x)
            else:
                retval = x.send(u)
            self.vm.return_value = retval
        except StopIteration as e:
            self.vm.pop()
            self.vm.push(e.value)
        else:
            # FIXME: The code has the effect of rerunning the last instruction.
            # I'm not sure if or why it is correct.
            if self.vm.version >= 3.6:
                self.vm.jump(self.vm.frame.f_lasti - 2)
            else:
                self.vm.jump(self.vm.frame.f_lasti - 1)
            return "yield"


if __name__ == "__main__":
    x = ByteOp33(None)



"""Bytecode Interpreter operations for Python 3.7
"""
from __future__ import print_function, division

import types
from xpython.byteop.byteop36 import ByteOp36

# Gone in 3.7
del ByteOp36.STORE_ANNOTATION
# del ByteOp36.WITH_CLEANUP_START
# del ByteOp36.WITH_CLEANUP_FINISH
# del ByteOp36.END_FINALLY
# del ByteOp36.POP_EXCEPT
# del ByteOp36.SETUP_WITH
# del ByteOp36.SETUP_ASYNC_WITH

class ByteOp37(ByteOp36):
    def __init__(self, vm):
        super(ByteOp37, self).__init__(vm)

    # Changed in 3.7

    # WITH_CLEANUP_START
    # WITH_CLEANUP_FINISH
    # END_FINALLY
    # POP_EXCEPT
    # SETUP_WITH
    # SETUP_ASYNC_WITH

    # New in 3.7

    ##############################################################################
    # Order of function here is the same as in:
    # https://docs.python.org/3.7/library/dis.htmls#python-bytecode-instructions
    #
    # A note about parameter names. Generally they are the same as
    # what is described above, however there are some slight changes:
    #
    # * when a parameter name is `namei` (an int), it appears as
    #   `name` (a str) below because the lookup on co_names[namei] has
    #   already been performed in parse_byte_and_args().
    ##############################################################################

    def LOAD_METHOD(self, name):
        """Loads a method named co_names[namei] from the TOS object. TOS is
        popped. This bytecode distinguishes two cases: if TOS has a
        method with the correct name, the bytecode pushes the unbound
        method and TOS. TOS will be used as the first argument (self)
        by CALL_METHOD when calling the unbound method. Otherwise,
        NULL and the object return by the attribute lookup are pushed.

        rocky: In our implementation in Python we don't have NULL: all
        stack entries have *some* value. So instead we'll push another
        item: the status. Also, instead of pushing the unbound method
        and self, we will pass the bound method, since that is what we
        have here. So TOS (self) is not pushed back onto the stack.
        """
        TOS = self.vm.pop()
        if hasattr(TOS, name):
            # FIXME: check that gettr(TO, name) is a method
            self.vm.push(getattr(TOS, name))
            self.vm.push("LOAD_METHOD lookup success")
        else:
            self.vm.push("fill in attribute method lookup")
            self.vm.push(None)

    def CALL_METHOD(self, count):
        """Calls a method. argc is the number of positional
        arguments. Keyword arguments are not supported. This opcode is
        designed to be used with LOAD_METHOD. Positional arguments are
        on top of the stack. Below them, the two items described in
        LOAD_METHOD are on the stack (either self and an unbound
        method object or NULL and an arbitrary callable). All of them
        are popped and the return value is pushed.

        rocky: In our setting, before "self" we have an additional
        item which is the status of the LOAD_METHOD. There is no way
        in Python to represent a value outside of a Python value which
        you can do in C, and is in effect what NULL is.
        """
        posargs = self.vm.popn(count)
        is_success = self.vm.pop()
        if is_success:
            func = self.vm.pop()
            self.call_function_with_args_resolved(func, posargs, {})
        else:
            # FIXME: do something else
            raise self.vm.PyVMError("CALL_METHOD not implemented yet")



"""Test the "with" statement for x-python."""

from __future__ import print_function

try:
    import vmtest
except ImportError:
    from . import vmtest

from xdis import PYTHON_VERSION

class TestWithStatement(vmtest.VmTestCase):

    def test_continue_in_with(self):
        self.self_checking()

    def test_simple_context_manager(self):
        self.self_checking()

    if PYTHON_VERSION >= 3.6:
        print("Test not gone over yet for >= 3.6")
    else:
        def test_raise_in_context_manager(self):
            self.do_one()

        if PYTHON_VERSION < 3.5:
            def test_suppressed_raise_in_context_manager(self):
                self.do_one()

        def test_return_in_with(self):
            self.assert_ok("""\
                class NullContext(object):
                    def __enter__(self):
                        l.append('i')
                        return self

                    def __exit__(self, exc_type, exc_val, exc_tb):
                        l.append('o')
                        return False

                l = []
                def use_with(val):
                    with NullContext():
                        l.append('w')
                        return val
                    l.append('e')

                assert use_with(23) == 23
                l.append('r')
                s = ''.join(l)
                print("Look: %r" % s)
                assert s == "iwor"
                """)

        def test_break_in_with(self):
            self.do_one()

        def test_raise_in_with(self):
            self.do_one()

        def test_at_context_manager_simplified(self):
            self.do_one()

        if PYTHON_VERSION >= 3.3:
            # FIXME - this a bug in our 3.2 handling
            def test_at_context_manager_complete(self):
                self.do_one()

            # "yield from" is not in 3.2-
            def test_generator_with_context_manager(self):
                self.do_one()



class GeneratorContextManager(object):
    def __init__(self, gen):
        self.gen = gen

    def __enter__(self):
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield")

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                next(self.gen)
            except StopIteration:
                return
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = type()
            try:
                self.gen.throw(type, value, traceback)
                raise RuntimeError(
                    "generator didn't stop after throw()"
                )
            except StopIteration as exc:
                return exc is not value
            except:
                if sys.exc_info()[1] is not value:
                    raise

def contextmanager(func):
    def helper(*args, **kwds):
        return GeneratorContextManager(func(*args, **kwds))
    return helper

@contextmanager
def my_context_manager(val):
    yield val

with my_context_manager(17) as x:
    assert x == 17



#!/usr/bin/env python
import re, sys
import os.path as osp
from xdis import IS_PYPY

# We do this crazy conversion from float to support Python 2.6 which
# doesn't support version_major, and has a bug in
# floating point so we can't divide 26 by 10 and get
# 2.6
PY_VERSION = sys.version_info[0] + (sys.version_info[1] / 10.0)

def get_srcdir():
    if PY_VERSION > 2.5:
        return osp.relpath(osp.normcase(osp.dirname(__file__)))
    else:
        return osp.normcase(osp.dirname(__file__))

if len(sys.argv) != 2:
    print("Usage: compile-file.py *byte-compiled-file*")
    sys.exit(1)

if PY_VERSION > 2.6:
    source = osp.normpath(osp.relpath(sys.argv[1]))
else:
    source = osp.normpath(sys.argv[1])

assert source.endswith('.py')
basename = osp.basename(source[:-3])

platform='pypy' if IS_PYPY else ''

bytecode_path = osp.normpath(osp.join(get_srcdir(),
                                      "bytecode-%s%s" % (platform, PY_VERSION),
                                      "%s.pyc" % basename))

import py_compile
print("compiling %s to %s" % (source, bytecode_path))
py_compile.compile(source, bytecode_path, source)
import os
os.system("xpython %s" % bytecode_path)



# Various functions

# Some Python's like 3.4 are irregular with the order in showing
# dictionary keys. For testing we need them to always come out in the
# same order.
def print_dict_sorted(mydict):
    for key, value in sorted(mydict.items(), key=lambda item: item[1]):
        print("\t%s: %s" % (key, value))
    print("")


def fn(*args):
    print("args is %r" % (args,))


fn(1, 2)


def fn(**kwargs):
    print("kwargs is")
    print_dict_sorted(kwargs)


fn(red=True, blue=False)


def fn(*args, **kwargs):
    print("args is %r" % (args,))
    print("kwargs is:")
    print_dict_sorted(kwargs)


fn(1, 2, red=True, blue=False)


def fn(x, y, *args, **kwargs):
    print("x is %r, y is %r" % (x, y))
    print("args is %r" % (args,))
    print("kwargs is")
    print_dict_sorted(kwargs)


fn("a", "b", 1, 2, red=True, blue=False)



# -*- coding: utf-8 -*-
"""Bytecode Interpreter operations base class

Note: this is subclassed. Later versions use operations from here.
"""
from __future__ import print_function, division

import inspect
import logging
import sys
from xdis import PYTHON_VERSION
from xpython.pyobj import Function
from xpython.buildclass import build_class

log = logging.getLogger(__name__)

class ByteOpBase(object):
    def __init__(self, vm):
        self.vm = vm
        # Convenience variables
        self.version = vm.version
        self.is_pypy = vm.is_pypy

        # Set this lazily in "convert_method_native_func
        self.method_func_access = None

    def build_container(self, count, container_fn):
        elts = self.vm.popn(count)
        self.vm.push(container_fn(elts))

    def call_function_with_args_resolved(self, func, posargs, namedargs):
        frame = self.vm.frame
        if hasattr(func, "im_func"):
            # Methods get self as an implicit first parameter.
            if func.im_self is not None:
                posargs.insert(0, func.im_self)
            # The first parameter must be the correct type.
            if not isinstance(posargs[0], func.im_class):
                raise TypeError(
                    "unbound method %s() must be called with %s instance "
                    "as first argument (got %s instance instead)"
                    % (
                        func.im_func.func_name,
                        func.im_class.__name__,
                        type(posargs[0]).__name__,
                    )
                )
            func = func.im_func

        # FIXME: should we special casing in a function?
        if inspect.isbuiltin(func):
            log.debug("handling built-in function %s" % func.__name__)
            if func == globals:
                # Use the frame's globals(), not the interpreter's
                self.vm.push(frame.f_globals)
                return
            elif func == locals:
                # Use the frame's locals(), not the interpreter's
                self.vm.push(frame.f_globals)
                return
            elif (self.is_pypy or self.version != PYTHON_VERSION) and PYTHON_VERSION >= 3.4:
                if func == __build_class__:
                    # later __build_class__() works only bytecode that matches the CPython interpeter,
                    # so use Darius' version instead.

                    # Try to convert to an interpreter function which is needed by build_class
                    if True: # self.version == PYTHON_VERSION and self.is_pypy != IS_PYPY:
                        assert len(posargs) > 0
                        posargs[0] = self.convert_native_to_Function(frame, posargs[0])

                    retval = build_class(self.vm.opc, *posargs, **namedargs)
                    self.vm.push(retval)
                    return

        if inspect.isfunction(func):
            # Try to convert to an interpreter function so we can interpret it.
            if func in self.vm.fn2native:
                func = self.vm.fn2native[func]
            elif False: # self.vm.version < 3.0:
                # Not quite ready. See 3.7 test_asyncgen.py for an
                # example of code that comes here. In that test, the
                # LOAD_GLOBAL '_ignore_deprecated_imports' fails to
                # find the global. '_ignore_deprecated_imports' is a method name
                # in test.support module of test/support/__init__.py.
                # In Python 2.X we work around a similar problem by
                # not tying to handle functions with closures.
                assert len(posargs) > 0
                posargs[0] = self.convert_native_to_Function(frame, posargs[0])

        if inspect.isfunction(func):
            log.debug("calling native function %s" % func.__name__)

        retval = func(*posargs, **namedargs)
        self.vm.push(retval)

    def call_function(self, argc, args, kwargs):
        namedargs = {}
        lenKw, lenPos = divmod(argc, 256)
        for i in range(lenKw):
            key, val = self.vm.popn(2)
            namedargs[key] = val
        namedargs.update(kwargs)
        posargs = self.vm.popn(lenPos)
        posargs.extend(args)

        func = self.vm.pop()
        self.call_function_with_args_resolved(func, posargs, namedargs)

    def convert_native_to_Function(self, frame, func):
        assert inspect.isfunction(func) or isinstance(func, Function)
        slots = {
            "kwdefaults" : {},
            "annotations": {},
        }
        if self.vm.version >= 3.0:
            slots["globs"] = frame.f_globals
            arg2attr = {
                "code": "__code__",
                "name": "__name__",
                "argdefs": "__defaults__",
                "kwdefaults": "__kwdefaults__",
                "annotations": "__annotations__",
                "closure": "__closure__",
                # FIXME: add __qualname__, __doc__
                # and __module__
            }
        else:
            slots["kwdefaults"] = {}
            slots["annotations"] = {}
            arg2attr = {
                "code": "func_code",
                "name": "__name__",
                "argdefs": "func_defaults",
                "globs": "func_globals",
                "annotations": "doesn't exist",
                "closure": "func_closure",
                # FIXME: add __doc__
                # and __module__
            }

        for argname, attribute in arg2attr.items():
            if hasattr(func, attribute):
                slots[argname] = getattr(func, attribute)

        closure = getattr(func, arg2attr["closure"])
        if not closure:
            # FIXME: we don't know how to convert functions with closures yet.
            native_func = func

            func = Function(
                slots["name"],
                slots["code"],
                slots["globs"],
                slots["argdefs"],
                slots["closure"],
                self.vm,
                slots["kwdefaults"],
                slots["annotations"],
            )
            self.vm.fn2native[native_func] = func
        return func

    def lookup_name(self, name):
        """Returns the value in the current frame associated for name"""
        frame = self.vm.frame
        if name in frame.f_locals:
            val = frame.f_locals[name]
        elif name in frame.f_globals:
            val = frame.f_globals[name]
        elif name in frame.f_builtins:
            val = frame.f_builtins[name]
        else:
            raise NameError("name '%s' is not defined" % name)
        return val

    def print_item(self, item, to=None):
        if to is None:
            to = sys.stdout

        # Python 2ish has file.softspace whereas
        # Python 3ish doesn't. Here is the doc on softspace:

        # Boolean that indicates whether a space character needs to be
        # printed before another value when using the print
        # statement. Classes that are trying to simulate a file object
        # should also have a writable softspace attribute, which
        # should be initialized to zero. This will be automatic for
        # most classes implemented in Python (care may be needed for
        # objects that override attribute access); types implemented
        # in C will have to provide a writable softspace attribute.

        # Note This attribute is not used to control the print
        # statement, but to allow the implementation of print to keep
        # track of its internal state.
        if hasattr(to, "softspace") and to.softspace:
            print(" ", end="", file=to)
            to.softspace = 0
        print(item, end="", file=to)

        if hasattr(to, "softspace"):
            if isinstance(item, str):
                if (not item) or (not item[-1].isspace()) or (item[-1] == " "):
                    to.softspace = 1
            else:
                to.softspace = 1

    def print_newline(self, to=None):
        if to is None:
            to = sys.stdout
        print("", file=to)
        if hasattr(to, "softspace"):
            to.softspace = 0



"""Bytecode Interpreter operations for Python 2.7
"""
from __future__ import print_function, division

import inspect
import types
from xpython.byteop.byteop25 import ByteOp25
from xpython.byteop.byteop26 import ByteOp26

# Gone since 2.6
del ByteOp25.JUMP_IF_FALSE
del ByteOp25.JUMP_IF_TRUE

class ByteOp27(ByteOp26):
    def __init__(self, vm):
        super(ByteOp27, self).__init__(vm)


    def convert_method_native_func(self, frame, method):
        """If a method's function is a native functions, converted it to the
        corresponding PyVM Method so that we can interpret it.
        """
        if not self.method_func_access:
            for func_attr in ("__func__", "im_func"):
                if hasattr(method, func_attr):
                    # Save attribute access name, so we don't
                    # have to compute this again.
                    self.method_func_access = func_attr
                    break
                pass
            else:
                raise self.vm.PyVMError("Can't find method function attribute; tried '__func__' and '_im_func'")
            pass

        func = getattr(method, self.method_func_access)
        if inspect.isfunction(func):
            func = self.convert_native_to_Function(self.vm.frame, func)
            method = types.MethodType(func, method.__self__)
        return method

    # New in 2.7

    # Note SET_ADD and MAP_ADD don't seem to be documented in
    # the 2.7 docs although the first appear there.
    # The docstring for these below is taken from 3.1 docs.
    # (3.0 doesn't have have MAP, although it has SET
    # which is exactly what is below.)
    def SET_ADD(self, count):
        """Calls set.add(TOS1[-i], TOS). Used to implement set
        comprehensions.
        """
        val = self.vm.pop()
        the_set = self.vm.peek(count)
        the_set.add(val)

    def MAP_ADD(self, count):
        """
        Calls dict.setitem(TOS1[-i], TOS, TOS1). Used to implement dict
        comprehensions.
        """
        val, key = self.vm.popn(2)
        the_map = self.vm.peek(count)
        the_map[key] = val

    # Note gone in 3.0 and 3.1, but appears again in 3.2
    def SETUP_WITH(self, delta):
        """
        This opcode performs several operations before a with block
        starts. First, it loads __exit__() from the context manager
        and pushes it onto the stack for later use by
        WITH_CLEANUP. Then, __enter__() is called, and a finally block
        pointing to delta is pushed. Finally, the result of calling
        the enter method is pushed onto the stack. The next opcode
        will either ignore it (POP_TOP), or store it in (a)
        variable(s) (STORE_FAST, STORE_NAME, or UNPACK_SEQUENCE).
        """
        context_manager = self.vm.pop()

        # Make sure __enter__ and __exit__ functions in context_manager are
        # converted to our Function type so we can interpret them.
        exit_method = self.convert_method_native_func(self.vm.frame, context_manager.__exit__)
        self.vm.push(exit_method)
        self.convert_method_native_func(self.vm.frame, context_manager.__enter__)
        finally_block = context_manager.__enter__()
        if self.version < 3.0:
            self.vm.push_block("with", delta)
        else:
            self.vm.push_block("finally", delta)
        self.vm.push(finally_block)

    def BUILD_SET(self, count):
        """Works as BUILD_TUPLE, but creates a set. New in version 2.7"""
        elts = self.vm.popn(count)
        self.vm.push(set(elts))

    def JUMP_FORWARD(self, delta):
        """Increments bytecode counter by delta."""
        self.vm.jump(delta)

    def POP_JUMP_IF_TRUE(self, target):
        """If TOS is true, sets the bytecode counter to target. TOS is popped."""
        val = self.vm.pop()
        if val:
            self.vm.jump(target)

    def POP_JUMP_IF_FALSE(self, target):
        """If TOS is false, sets the bytecode counter to target. TOS is popped."""
        val = self.vm.pop()
        if not val:
            self.vm.jump(target)

    def JUMP_IF_TRUE_OR_POP(self, target):
        """
        If TOS is true, sets the bytecode counter to target and leaves TOS
        on the stack. Otherwise (TOS is false), TOS is popped.
        """
        val = self.vm.top()
        if val:
            self.vm.jump(target)
        else:
            self.vm.pop()

    def JUMP_IF_FALSE_OR_POP(self, target):
        """
        If TOS is false, sets the bytecode counter to target and leaves TOS
        on the stack. Otherwise (TOS is true), TOS is popped.
        """
        val = self.vm.top()
        if not val:
            self.vm.jump(target)
        else:
            self.vm.pop()

    def JUMP_ABSOLUTE(self, target):
        self.vm.jump(target)



# -*- coding: utf-8 -*-
"""Byte Interpreter operations for Python 3.2
"""
from __future__ import print_function, division


from xpython.byteop.byteop27 import ByteOp27
from xpython.byteop.byteop25 import ByteOp25
from xpython.pyobj import Function

# FIXME: investigate does "del" removing and attribute here
# have an effect on what another module sees as ByteOp27's attributes?

# Gone since 3.0
del ByteOp25.PRINT_EXPR
del ByteOp25.PRINT_ITEM
del ByteOp25.PRINT_ITEM_TO
del ByteOp25.PRINT_NEWLINE
del ByteOp25.PRINT_NEWLINE_TO
del ByteOp25.BUILD_CLASS
del ByteOp25.EXEC_STMT
del ByteOp25.RAISE_VARARGS

# Gone since 3.2
del ByteOp25.DUP_TOPX

# Code with these names have an implicit .0 in them
COMPREHENSION_FN_NAMES = frozenset((
    "<setcomp>",
    "<dictcomp>",
    "<genexpr>",
))


class ByteOp32(ByteOp27):
    def __init__(self, vm):
        super(ByteOp32, self).__init__(vm)

    def do_raise(self, exc, cause):
        if exc is None:  # reraise
            exc_type, val, tb = self.vm.last_exception
            if exc_type is None:
                return "exception"  # error
            else:
                return "reraise"

        elif type(exc) == type:
            # As in `raise ValueError`
            exc_type = exc
            val = exc()  # Make an instance.
        elif isinstance(exc, BaseException):
            # As in `raise ValueError('foo')`
            exc_type = type(exc)
            val = exc
        else:
            return "exception"  # error

        # If you reach this point, you're guaranteed that
        # val is a valid exception instance and exc_type is its class.
        # Now do a similar thing for the cause, if present.
        if cause:
            if type(cause) == type:
                cause = cause()
            elif not isinstance(cause, BaseException):
                return "exception"  # error

            val.__cause__ = cause

        self.vm.last_exception = exc_type, val, val.__traceback__
        return "exception"

    # Changed from 2.7
    # 3.2 has kwdefaults that aren't allowed in 2.7
    def MAKE_FUNCTION(self, argc):
        """
        Pushes a new function object on the stack. From bottom to top, the consumed stack must consist of:

        * argc & 0xFF default argument objects in positional order
        * (argc >> 8) & 0xFF pairs of name and default argument, with the name just below the object on the stack, for keyword-only parameters
        * (argc >> 16) & 0x7FFF parameter annotation objects
        * a tuple listing the parameter names for the annotations (only if there are ony annotation objects)
        * the code associated with the function (at TOS1 if 3.3+ else at TOS for 3.0..3.2)
        * the qualified name of the function (at TOS if 3.3+)
        """
        rest, default_count = divmod(argc, 256)
        annotate_count, kw_default_count = divmod(rest, 256)

        if self.version >= 3.3:
            name = self.vm.pop()
            code = self.vm.pop()
        else:
            code = self.vm.pop()
            name = code.co_name

        if annotate_count:
            annotate_names = self.vm.pop()
            # annotate count includes +1 for the above names
            annotate_objects = self.vm.popn(annotate_count - 1)
            n = len(annotate_names)
            assert n == len(annotate_objects)
            annotations = {annotate_names[i]: annotate_objects[i] for i in range(n)}
        else:
            annotations = {}

        if kw_default_count:
            kw_default_pairs = self.vm.popn(2 * kw_default_count)
            kwdefaults = dict(
                kw_default_pairs[i : i + 2] for i in range(0, len(kw_default_pairs), 2)
            )
        else:
            kwdefaults = {}

        if default_count:
            defaults = self.vm.popn(default_count)
        else:
            defaults = tuple()

        globs = self.vm.frame.f_globals

        fn = Function(
            name,
            code,
            globs,
            defaults,
            closure=None,
            vm=self.vm,
            kwdefaults=kwdefaults,
            annotations=annotations,
        )

        if argc == 0 and name in COMPREHENSION_FN_NAMES:
            fn.has_dot_zero = True

        fn.version = self.version
        self.vm.push(fn)

    # Order of function here is the same as in:
    # https://docs.python.org/3.2/library/dis.html#python-bytecode-instructions

    # Note these are only the functions that aren't in the parent class
    # here, Python 2.7

    def DUP_TOP_TWO(self):
        """Duplicates the reference on top of the stack."""
        a, b = self.vm.popn(2)
        self.vm.push(a, b, a, b)

    def POP_EXCEPT(self):
        """
        Removes one block from the block stack. The popped block must be an
        exception handler block, as implicitly created when entering an except
        handler. In addition to popping extraneous values from the frame
        stack, the last three popped values are used to restore the exception
        state."""
        block = self.vm.pop_block()
        if block.type != "except-handler":
            raise self.vm.PyVMError(
                "popped block is not an except handler; is %s" % block
            )
        self.vm.unwind_block(block)

    def LOAD_BUILD_CLASS(self):
        """Pushes builtins.__build_class__() onto the stack. It is later called by CALL_FUNCTION to construct a class."""
        self.vm.push(__build_class__)

    # This opcode disappears starting in 3.5
    def WITH_CLEANUP(self):
        """Cleans up the stack when a `with` statement block exits. TOS is the
        context manager's `__exit__()` bound method.

        Below TOS are 1 3 values indicating how/why the finally clause
        was entered:

        * SECOND = None
        * (SECOND, THIRD) = (WHY_{RETURN,CONTINUE}), retval
        * SECOND = WHY_*; no retval below it
        * (SECOND, THIRD, FOURTH) = exc_info()

        In the last case, EXIT(SECOND, THIRD, FOURTH) is called,
        otherwise TOS(None, None, None). In addition, TOS is removed from the stack.

        If the stack represents an exception, and the function call
        returns a  true  value, this information is  zapped  and
        replaced with a single WHY_SILENCED to prevent END_FINALLY
        from re-raising the exception. (But non-local gotos will still
        be resumed.)
        """
        # Note: the code is derived from byterun where it had to handle
        # both 2.7 and 3.3.
        v = w = None
        u = self.vm.top()
        if u is None:
            exit_func = self.vm.pop(1)
        elif isinstance(u, str):
            if u in ("return", "continue"):
                exit_func = self.vm.pop(2)
            else:
                exit_func = self.vm.pop(1)
            u = None
        elif issubclass(u, BaseException):
            w, v, u = self.vm.popn(3)
            tp, exc, tb = self.vm.popn(3)
            exit_func = self.vm.pop()
            self.vm.push(tp, exc, tb)
            self.vm.push(None)
            self.vm.push(w, v, u)
            block = self.vm.pop_block()
            assert block.type == "except-handler"
            self.vm.push_block(block.type, block.handler, block.level - 1)
        else:  # pragma: no cover
            raise self.vm.PyVMError("Confused WITH_CLEANUP")
        exit_ret = exit_func(u, v, w)
        err = (u is not None) and bool(exit_ret)
        if err:
            # An error occurred, and was suppressed
            self.vm.push("silenced")

    # Note: this is gone in 3.4
    def STORE_LOCALS(self):
        """Pops TOS from the stack and stores it as the current frames f_locals. This is used in class construction."""
        self.vm.frame.f_locals = self.vm.pop()

    def RAISE_VARARGS(self, argc):
        """
        Raises an exception. argc indicates the number of arguments to the
        raise statement, ranging from 0 to 3. The handler will find
        the traceback as TOS2, the parameter as TOS1, and the
        exception as TOS.
        """
        cause = exc = None
        if argc == 2:
            cause = self.vm.pop()
            exc = self.vm.pop()
        elif argc == 1:
            exc = self.vm.pop()
        return self.do_raise(exc, cause)


if __name__ == "__main__":
    x = ByteOp32(None)



from xpython.interpreter.baseobjspace import W_Root
from xpython.interpreter.error import OperationError, oefmt
from xpython.interpreter.pyopcode import SApplicationException
from xdis import CO_COROUTINE, CO_ITERABLE_COROUTINE, CO_ASYNC_GENERATOR, PYPY_COMPILER_FLAG_NAMES

class GeneratorOrCoroutine(W_Root):
    _immutable_fields_ = ['pycode']

    def __init__(self, frame, name=None, qualname=None):
        self.space = frame.space
        self.frame = frame     # turned into None when frame_finished_execution
        self.pycode = frame.pycode  # usually never None but see __setstate__
        self.running = False
        self._name = name           # may be null, use get_name()
        self._qualname = qualname   # may be null, use get_qualname()
        if (isinstance(self, Coroutine)
                or self.pycode.co_flags & PYPY_COMPILER_FLAG_NAMES["CO_YIELD_INSIDE_TRY"]):
            self.register_finalizer(self.space)
        self.saved_operr = None

    def get_name(self):
        # 'name' is a byte string that is valid utf-8
        if self._name is not None:
            return self._name
        elif self.pycode is None:
            return "<finished>"
        else:
            return self.pycode.co_name

    def get_qualname(self):
        # 'qualname' is a unicode string
        if self._qualname is not None:
            return self._qualname
        return self.get_name()

    def descr__repr__(self, space):
        addrstring = self.getaddrstring(space)
        return space.newtext("<%s object %s at 0x%s>" %
                          (self.KIND, self.get_qualname(), addrstring))

    def descr_send(self, w_arg):
        """send(arg) -> send 'arg' into generator/coroutine,
return next yielded value or raise StopIteration."""
        return self.send_ex(w_arg)

    def send_error(self, operr):
        return self.send_ex(SApplicationException(operr))

    def send_ex(self, w_arg_or_err):
        assert w_arg_or_err is not None
        pycode = self.pycode
        assert pycode is not None
        return self._send_ex(w_arg_or_err)

    def _send_ex(self, w_arg_or_err):
        space = self.space

        frame = self.frame
        if frame is None:
            if isinstance(self, Coroutine):
                # NB. CPython checks a flag 'closing' here, but instead
                # we can simply not be here at all if frame is None in
                # this case
                raise RuntimeError("cannot reuse already awaited coroutine")
            # xxx a bit ad-hoc, but we don't want to go inside
            # execute_frame() if the frame is actually finished
            if isinstance(w_arg_or_err, SApplicationException):
                operr = w_arg_or_err.operr
            elif isinstance(self, AsyncGenerator):
                operr = OperationError(space.w_StopAsyncIteration, space.w_None)
            else:
                operr = OperationError(space.w_StopIteration, space.w_None)
            raise operr

        w_result = self._invoke_execute_frame(w_arg_or_err)
        assert w_result is not None

        # if the frame is now marked as finished, it was RETURNed from
        if frame.frame_finished_execution:
            self.frame_is_finished()
            if isinstance(self, AsyncGenerator):
                assert space.is_w(w_result, space.w_None), (
                    "getting non-None here should be forbidden by the bytecode")
                raise OperationError(space.w_StopAsyncIteration, space.w_None)
            else:
                if space.is_w(w_result, space.w_None):
                    raise OperationError(space.w_StopIteration, space.w_None)
                else:
                    raise stopiteration_value(space, w_result)
        else:
            return w_result     # YIELDed

    def _invoke_execute_frame(self, w_arg_or_err):
        space = self.space
        frame = self.frame
        if self.running:
            raise oefmt(space.w_ValueError, "%s already executing", self.KIND)
        #
        # Optimization only: after we've started a Coroutine without
        # CO_YIELD_INSIDE_TRY, then Coroutine._finalize_() will be a no-op
        if frame.last_instr == -1:
            if (not space.is_w(w_arg_or_err, space.w_None) and
                not isinstance(w_arg_or_err, SApplicationException)):
                raise oefmt(space.w_TypeError,
                            "can't send non-None value to a just-started %s",
                            self.KIND)
        ec = space.getexecutioncontext()
        current_exc_info = ec.sys_exc_info()
        if self.saved_operr is not None:
            ec.set_sys_exc_info(self.saved_operr)
            self.saved_operr = None
        self.running = True
        try:
            w_result = frame.execute_frame(w_arg_or_err)
        except OperationError as e:
            # errors finish a frame
            try:
                if e.match(space, space.w_StopIteration):
                    self._leak_stopiteration(e)
                elif (isinstance(self, AsyncGenerator) and
                      e.match(space, space.w_StopAsyncIteration)):
                    self._leak_stopasynciteration(e)
            finally:
                self.frame_is_finished()
            raise
        finally:
            self.running = False
            # note: this is not perfectly correct: see
            # test_exc_info_in_generator_4.  But it's simpler and
            # bug-to-bug compatible with CPython 3.5 and 3.6.
            if frame._any_except_or_finally_handler():
                self.saved_operr = ec.sys_exc_info()
            ec.set_sys_exc_info(current_exc_info)
        return w_result

    def get_delegate(self):
        if self.frame is None:
            return None
        return self.frame.w_yielding_from

    def descr_delegate(self, space):
        w_yf = self.get_delegate()
        if w_yf is None:
            return space.w_None
        return w_yf

    def set_delegate(self, w_delegate):
        self.frame.w_yielding_from = w_delegate

    def _leak_stopiteration(self, e):
        # Check for __future__ generator_stop and conditionally turn
        # a leaking StopIteration into RuntimeError (with its cause
        # set appropriately).
        space = self.space
        if (
            self.pycode.co_flags
                & (
                    CO_COROUTINE |
                    CO_ITERABLE_COROUTINE |
                    CO_ASYNC_GENERATOR
                    # | CO_FUTURE_GENERATOR_STOP # after next xdis release
                )):
            e2 = OperationError(space.w_RuntimeError,
                                space.newtext("%s raised StopIteration" %
                                              self.KIND))
            e2.chain_exceptions_from_cause(space, e)
            raise e2
        else:
            space.warn(space.newtext("generator '%s' raised StopIteration"
                                        % self.get_qualname()),
                       space.w_DeprecationWarning)

    def _leak_stopasynciteration(self, e):
        space = self.space
        e2 = OperationError(space.w_RuntimeError,
                   space.newtext("async generator raised StopAsyncIteration"))
        e2.chain_exceptions_from_cause(space, e)
        raise e2

    def descr_throw(self, w_type, w_val=None, w_tb=None):
        """throw(typ[,val[,tb]]) -> raise exception in generator/coroutine,
return next yielded value or raise StopIteration."""
        return self.throw(w_type, w_val, w_tb)

    def throw(self, w_type, w_val, w_tb):
        from pypy.interpreter.pytraceback import check_traceback

        space = self.space
        if w_val is None:
            w_val = space.w_None

        msg = "throw() third argument must be a traceback object"
        if space.is_none(w_tb):
            tb = None
        else:
            tb = check_traceback(space, w_tb, msg)

        operr = OperationError(w_type, w_val, tb)
        operr.normalize_exception(space)

        # note: _w_yielded_from is always None if 'self.running'
        if (self.get_delegate() is not None and
                    operr.match(space, space.w_GeneratorExit)):
            try:
                self._gen_close_iter(space)
            except OperationError as e:
                return self.send_error(e)

        if tb is None:
            tb = space.getattr(operr.get_w_value(space),
                               space.newtext('__traceback__'))
            if not space.is_w(tb, space.w_None):
                operr.set_traceback(tb)
        return self.send_error(operr)

    def _gen_close_iter(self, space):
        assert not self.running
        w_yf = self.get_delegate()
        self.set_delegate(None)
        self.running = True
        try:
            gen_close_iter(space, w_yf)
        finally:
            self.running = False

    def descr_close(self):
        """close() -> raise GeneratorExit inside generator/coroutine."""
        if self.frame is None:
            return     # nothing to do in this case
        space = self.space
        operr = None
        # note: _w_yielded_from is always None if 'self.running'
        w_yf = self.get_delegate()
        if w_yf is not None:
            try:
                self._gen_close_iter(space)
            except OperationError as e:
                operr = e
        if operr is None:
            operr = OperationError(space.w_GeneratorExit, space.w_None)
        try:
            self.send_error(operr)
        except OperationError as e:
            if e.match(space, space.w_StopIteration) or \
                    e.match(space, space.w_GeneratorExit):
                return space.w_None
            raise
        else:
            raise oefmt(space.w_RuntimeError,
                        "%s ignored GeneratorExit", self.KIND)

    def descr_gicr_frame(self, space):
        if self.frame is not None and not self.frame.frame_finished_execution:
            return self.frame
        else:
            return space.w_None

    def descr__name__(self, space):
        return space.newtext(self.get_name())

    def descr_set__name__(self, space, w_name):
        if space.isinstance_w(w_name, space.w_unicode):
            self._name = space.text_w(w_name)
        else:
            raise oefmt(space.w_TypeError,
                        "__name__ must be set to a string object")

    def descr__qualname__(self, space):
        return space.newtext(self.get_qualname())

    def descr_set__qualname__(self, space, w_name):
        try:
            self._qualname = space.utf8_w(w_name)
        except OperationError as e:
            if e.match(space, space.w_TypeError):
                raise oefmt(space.w_TypeError,
                            "__qualname__ must be set to a string object")
            raise

    def _finalize_(self):
        # This is only called if the CO_YIELD_INSIDE_TRY flag is set
        # on the code object.  If the frame is still not finished and
        # finally or except blocks are present at the current
        # position, then raise a GeneratorExit.  Otherwise, there is
        # no point.
        if self.frame is not None:
            block = self.frame.lastblock
            while block is not None:
                if not isinstance(block, LoopBlock):
                    self.descr_close()
                    break
                block = block.previous

    def frame_is_finished(self):
        self.frame = None
        rgc.may_ignore_finalizer(self)

    def iterator_greenkey(self, space):
        return self.pycode


class GeneratorIterator(GeneratorOrCoroutine):
    "An iterator created by a generator."
    KIND = "generator"
    KIND_U = u"generator"

    def descr__iter__(self):
        """Implement iter(self)."""
        return self

    def descr_next(self):
        """Implement next(self)."""
        return self.send_ex(self.space.w_None)

    # Results can be either an RPython list of W_Root, or it can be an
    # app-level W_ListObject, which also has an append() method, that's why we
    # generate 2 versions of the function and 2 jit drivers.
    def _create_unpack_into():

        def unpack_into(self, results):
            """This is a hack for performance: runs the generator and collects
            all produced items in a list."""
            frame = self.frame
            if frame is None:    # already finished
                return
            pycode = self.pycode
            while True:
                space = self.space
                try:
                    w_result = self._invoke_execute_frame(space.w_None)
                except OperationError as e:
                    if not e.match(space, space.w_StopIteration):
                        raise
                    break
                # if the frame is now marked as finished, it was RETURNed from
                if frame.frame_finished_execution:
                    self.frame_is_finished()
                    break
                results.append(w_result)     # YIELDed
        return unpack_into
    unpack_into = _create_unpack_into()
    unpack_into_w = _create_unpack_into()


class Coroutine(GeneratorOrCoroutine):
    "A coroutine object."
    KIND = "coroutine"
    KIND_U = u"coroutine"

    def descr__await__(self, space):
        return CoroutineWrapper(self)

    def _finalize_(self):
        # If coroutine was never awaited on issue a RuntimeWarning.
        if self.pycode is not None and \
           self.frame is not None and \
           self.frame.last_instr == -1:
            space = self.space
            msg = "coroutine '%s' was never awaited" % self.get_qualname()
            space.warn(space.newtext(msg), space.w_RuntimeWarning)
        GeneratorOrCoroutine._finalize_(self)


class CoroutineWrapper(W_Root):
    _immutable_ = True

    def __init__(self, coroutine):
        self.coroutine = coroutine

    def descr__iter__(self, space):
        return self

    def descr__next__(self, space):
        return self.coroutine.send_ex(space.w_None)

    def descr_send(self, space, w_arg):
        return self.coroutine.send_ex(w_arg)
    descr_send.__doc__ = Coroutine.descr_send.__doc__

    def descr_throw(self, w_type, w_val=None, w_tb=None):
        return self.coroutine.throw(w_type, w_val, w_tb)
    descr_throw.__doc__ = Coroutine.descr_throw.__doc__

    def descr_close(self):
        return self.coroutine.descr_close()
    descr_close.__doc__ = Coroutine.descr_close.__doc__


class AIterWrapper(W_Root):
    # NB. this type was added in CPython 3.5.2
    _immutable_ = True

    def __init__(self, w_aiter):
        self.w_aiter = w_aiter

    def descr__await__(self, space):
        return self

    def descr__iter__(self, space):
        return self

    def descr__next__(self, space):
        raise stopiteration_value(space, self.w_aiter)

def stopiteration_value(space, w_value):
    # Mess.  The obvious line, "OperationError(w_StopIteration, w_value)",
    # fails for some types of w_value.  E.g. if it's a subclass of
    # tuple, it will be unpacked as individual arguments.
    raise OperationError(space.w_StopIteration,
                         space.call_function(space.w_StopIteration, w_value))


def gen_close_iter(space, w_yf):
    # This helper function is used by close() and throw() to
    # close a subiterator being delegated to by yield-from.
    if isinstance(w_yf, GeneratorIterator) or isinstance(w_yf, Coroutine):
        w_yf.descr_close()
    else:
        try:
            w_close = space.getattr(w_yf, space.newtext("close"))
        except OperationError as e:
            if not e.match(space, space.w_AttributeError):
                # aaaaaaaah but that's what CPython does too
                e.write_unraisable(space, "generator/coroutine.close()")
        else:
            space.call_function(w_close)

def gen_is_coroutine(w_obj):
    return (isinstance(w_obj, GeneratorIterator) and
            (w_obj.pycode.co_flags & CO_ITERABLE_COROUTINE) != 0)

def get_awaitable_iter(space, w_obj):
    # This helper function returns an awaitable for `o`:
    #    - `o` if `o` is a coroutine-object;
    #    - otherwise, o.__await__()

    if isinstance(w_obj, Coroutine) or gen_is_coroutine(w_obj):
        return w_obj

    w_await = space.lookup(w_obj, "__await__")
    if w_await is None:
        raise oefmt(space.w_TypeError,
                    "object %T can't be used in 'await' expression",
                    w_obj)
    w_res = space.get_and_call_function(w_await, w_obj)
    if isinstance(w_res, Coroutine) or gen_is_coroutine(w_res):
        raise oefmt(space.w_TypeError,
                    "__await__() returned a coroutine (it must return an "
                    "iterator instead, see PEP 492)")
    elif space.lookup(w_res, "__next__") is None:
        raise oefmt(space.w_TypeError,
                "__await__() returned non-iterator "
                "of type '%T'", w_res)
    return w_res


# ----------


def get_printable_location_genentry(bytecode):
    return '%s <generator>' % (bytecode.get_repr(),)

# ------------------------------------------------
# Python 3.6 async generators


class AsyncGenerator(GeneratorOrCoroutine):
    "An async generator (i.e. a coroutine with a 'yield')"
    KIND = "async generator"
    KIND_U = u"async_generator"

    def __init__(self, frame, name=None, qualname=None):
        self.hooks_inited = False
        GeneratorOrCoroutine.__init__(self, frame, name, qualname)

    def init_hooks(self):
        if self.hooks_inited:
            return
        self.hooks_inited = True
        ec = self.space.getexecutioncontext()
        self.w_finalizer = ec.w_asyncgen_finalizer_fn
        w_firstiter = ec.w_asyncgen_firstiter_fn
        if w_firstiter is not None:
            self.space.call_function(w_firstiter, self)

    def _finalize_(self):
        if self.frame is not None and self.frame.lastblock is not None:
            if self.w_finalizer is not None:
                try:
                    self.space.call_function(self.w_finalizer, self)
                except OperationError as e:
                    e.write_unraisable(self.space, "async generator finalizer")
                return
        GeneratorOrCoroutine._finalize_(self)

    def descr__aiter__(self):
        """Return an asynchronous iterator."""
        return self

    def descr__anext__(self):
        self.init_hooks()
        return AsyncGenASend(self, self.space.w_None)

    def descr_asend(self, w_arg):
        self.init_hooks()
        return AsyncGenASend(self, w_arg)

    def descr_athrow(self, w_type, w_val=None, w_tb=None):
        self.init_hooks()
        return AsyncGenAThrow(self, w_type, w_val, w_tb)

    def descr_aclose(self):
        self.init_hooks()
        return AsyncGenAThrow(self, None, None, None)


class AsyncGenValueWrapper(W_Root):
    def __init__(self, w_value):
        self.w_value = w_value


class AsyncGenABase(W_Root):
    ST_INIT = 0
    ST_ITER = 1
    ST_CLOSED = 2

    state = ST_INIT

    def __init__(self, async_gen):
        self.space = async_gen.space
        self.async_gen = async_gen

    def descr__iter__(self):
        return self

    def descr__next__(self):
        return self.do_send(self.space.w_None)

    def descr_send(self, w_arg):
        return self.do_send(w_arg)

    def descr_throw(self, w_type, w_val=None, w_tb=None):
        return self.do_throw(w_type, w_val, w_tb)

    def descr_close(self):
        self.state = self.ST_CLOSED

    def unwrap_value(self, w_value):
        if isinstance(w_value, AsyncGenValueWrapper):
            w_value = self.space.call_function(self.space.w_StopIteration,
                                               w_value.w_value)
            raise OperationError(self.space.w_StopIteration, w_value)
        else:
            return w_value


class AsyncGenASend(AsyncGenABase):

    def __init__(self, async_gen, w_value_to_send):
        AsyncGenABase.__init__(self, async_gen)
        self.w_value_to_send = w_value_to_send

    def do_send(self, w_arg_or_err):
        space = self.space
        if self.state == self.ST_CLOSED:
            raise OperationError(space.w_StopIteration, space.w_None)

        # We think that the code should look like this:
        #if self.w_value_to_send is not None:
        #    if not space.is_w(w_arg_or_err, space.w_None):
        #        raise ...
        #    w_arg_or_err = self.w_value_to_send
        #    self.w_value_to_send = None

        # But instead, CPython's logic is this, which we think is
        # giving nonsense results for 'g.asend(42).send(43)':
        if self.state == self.ST_INIT:
            if space.is_w(w_arg_or_err, space.w_None):
                w_arg_or_err = self.w_value_to_send
            self.state = self.ST_ITER

        try:
            w_value = self.async_gen.send_ex(w_arg_or_err)
            return self.unwrap_value(w_value)
        except OperationError:
            self.state = self.ST_CLOSED
            raise

    def do_throw(self, w_type, w_val, w_tb):
        space = self.space
        if self.state == self.ST_CLOSED:
            raise OperationError(space.w_StopIteration, space.w_None)
        try:
            w_value = self.async_gen.throw(w_type, w_val, w_tb)
            return self.unwrap_value(w_value)
        except OperationError:
            self.state = self.ST_CLOSED
            raise


class AsyncGenAThrow(AsyncGenABase):

    def __init__(self, async_gen, w_exc_type, w_exc_value, w_exc_tb):
        AsyncGenABase.__init__(self, async_gen)
        self.w_exc_type = w_exc_type
        self.w_exc_value = w_exc_value
        self.w_exc_tb = w_exc_tb

    def do_send(self, w_arg_or_err):
        # XXX FAR MORE COMPLICATED IN CPYTHON
        space = self.space
        if self.state == self.ST_CLOSED:
            raise OperationError(space.w_StopIteration, space.w_None)

        if self.state == self.ST_INIT:
            if not space.is_w(w_arg_or_err, space.w_None):
                raise OperationError(space.w_RuntimeError, space.newtext(
                    "can't send non-None value to a just-started coroutine"))
            self.state = self.ST_ITER
            throwing = True
        else:
            throwing = False

        try:
            if throwing:
                if self.w_exc_type is None:
                    # TODO: add equivalent to CPython's o->agt_gen->ag_closed = 1;
                    w_value = self.async_gen.throw(space.w_GeneratorExit,
                                                   None, None)
                else:
                    w_value = self.async_gen.throw(self.w_exc_type,
                                                   self.w_exc_value,
                                                   self.w_exc_tb)
            else:
                w_value = self.async_gen.send_ex(w_arg_or_err)

            if (self.w_exc_type is None and w_value is not None and
                    isinstance(w_value, AsyncGenValueWrapper)):
                raise oefmt(space.w_RuntimeError,
                            "async generator ignored GeneratorExit")
            return self.unwrap_value(w_value)
        except OperationError as e:
            self.handle_error(e)

    def do_throw(self, w_type, w_val, w_tb):
        space = self.space
        if self.state == self.ST_INIT:
            raise OperationError(self.space.w_RuntimeError,
                space.newtext("can't do async_generator.athrow().throw()"))
        if self.state == self.ST_CLOSED:
            raise OperationError(space.w_StopIteration, space.w_None)
        try:
            w_value = self.async_gen.throw(w_type, w_val, w_tb)
            return self.unwrap_value(w_value)
        except OperationError as e:
            self.handle_error(e)

    def handle_error(self, e):
        space = self.space
        self.state = self.ST_CLOSED
        if (e.match(space, space.w_StopAsyncIteration) or
                    e.match(space, space.w_StopIteration)):
            if self.w_exc_type is None:
                # When aclose() is called we don't want to propagate
                # StopAsyncIteration or GeneratorExit; just raise
                # StopIteration, signalling that this 'aclose()' await
                # is done.
                raise OperationError(space.w_StopIteration, space.w_None)
        if e.match(space, space.w_GeneratorExit):
            if self.w_exc_type is None:
                # Ignore this error.
                raise OperationError(space.w_StopIteration, space.w_None)
        raise e



class NullContext(object):
    def __enter__(self):
        l.append("i")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert exc_type is ValueError, "Expected ValueError: %r" % exc_type
        l.append("o")
        return False


l = []
try:
    with NullContext():
        l.append("w")
        raise ValueError("Boo!")
    l.append("e")
except ValueError:
    l.append("x")
l.append("r")
s = "".join(l)
print("Look: %r" % s)
assert s == "iwoxr"



# Inplace operator tests
"""This program is self-checking!"""

x, y = 2, 3
x **= y
assert x == 8 and y == 3
x *= y
assert x == 24 and y == 3
x //= y
assert x == 8 and y == 3
x %= y
assert x == 2 and y == 3
x += y
assert x == 5 and y == 3
x -= y
assert x == 2 and y == 3
x <<= y
assert x == 16 and y == 3
x >>= y
assert x == 2 and y == 3

x = 0x8F
x &= 0xA5
assert x == 0x85
x |= 0x10
assert x == 0x95
x ^= 0x33
assert x == 0xA6



# Test of 3.5+ GET_YIELD_FROM_ITER

# Code is from https://stackoverflow.com/questions/41136410/python-yield-from-or-return-a-generator
def add(a, b):
    return a + b

def sqrt(a):
    return a ** 0.5

data1 = [*zip(range(1, 3))]  # [(1,), (2,)]

job1 = (sqrt, data1)

def gen_factory(func, seq):
    """Generator factory returning a generator."""
    # do stuff ... immediately when factory gets called
    print("build generator & return")
    return (func(*args) for args in seq)

def gen_generator(func, seq):
    """Generator yielding from sub-generator inside."""
    # do stuff ... first time when 'next' gets called
    print("build generator & yield")
    yield from (func(*args) for args in seq)

gen_fac = gen_factory(*job1)
print(gen_fac)
# build generator & return <-- printed immediately
print(next(gen_fac))  # start
# Out: 1.0
print([*gen_fac])  # deplete rest of generator
# Out: [1.4142135623730951]



# Test slice assignment
l = list(range(6))

l[2:5] = ["x"]
assert l == [0, 1, 'x', 5]

l = list(range(7))
l[:5] = ["x"]
assert l == ['x', 5, 6]

l = list(range(6))
l[2:] = ["x"]
assert l == [0, 1, 'x']

l = list(range(6))
l[:] = ["x"]
assert l == ['x']

# Test slice deletion

l = list(range(6))
del l[2:5]
assert l == [0, 1, 5]

l = list(range(6))
del l[:5]
assert l == [5]

l = list(range(6))
del l[2:]
assert l == [0, 1]

l = list(range(6))
del l[:]
assert l == []

l = list(range(6))
del l[::2]
assert l == [1, 3, 5]



"""Execute files of Python code."""

import os
import sys
import tokenize
import mimetypes
from xdis import load_module, PYTHON_VERSION, IS_PYPY

from xpython.vm import PyVM, PyVMUncaughtException
from xpython.vmtrace import PyVMTraced
from xpython.version import SUPPORTED_PYTHON, SUPPORTED_BYTECODE, SUPPORTED_PYPY

# To silence the "import imp" DeprecationWarning below
import warnings

warnings.filterwarnings("ignore")
import imp

# This code is ripped off from coverage.py.  Define things it expects.
try:
    open_source = tokenize.open  # pylint: disable=E1101
except:

    def open_source(fname):
        """Open a source file the best way."""
        return open(fname, "rU")


class CannotCompileError(Exception):
    """For raising errors when we have a Compile eror."""

    pass


class WrongBytecodeError(Exception):
    """For raising errors when we have the wrong bytecode."""

    pass


class NoSourceError(Exception):
    """For raising errors when we can't find source code."""

    pass


def exec_code_object(
    code, env, python_version=PYTHON_VERSION, is_pypy=IS_PYPY, callback=None
):
    if callback:
        vm = PyVMTraced(callback, python_version, is_pypy)
        try:
            vm.run_code(code, f_globals=env)
        except PyVMUncaughtException:
            vm.last_exception = event_arg = (
                vm.last_exception[0],
                vm.last_exception[1],
                vm.last_traceback)
            callback("fatal", 0, "fatalOpcode", 0, -1, event_arg, vm)
    else:
        vm = PyVM(python_version, is_pypy)
        try:
            vm.run_code(code, f_globals=env)
        except PyVMUncaughtException:
            pass

def get_supported_versions(is_pypy, is_bytecode):
    if is_bytecode:
        supported_versions = SUPPORTED_PYPY if IS_PYPY else SUPPORTED_BYTECODE
        mess = (
            "PYPY 2.7, 3.2, 3.5 and 3.6"
            if is_pypy
            else "CPython 2.5 .. 2.7, 3.2 .. 3.7"
        )
    else:
        supported_versions = SUPPORTED_PYPY if IS_PYPY else SUPPORTED_PYTHON
        mess = "PYPY 2.7, 3.2, 3.5 and 3.6" if is_pypy else "CPython 2.7, 3.2 .. 3.7"
    return supported_versions, mess


# from coverage.py:

try:
    # In Py 2.x, the builtins were in __builtin__
    BUILTINS = sys.modules["__builtin__"]
except KeyError:
    # In Py 3.x, they're in builtins
    BUILTINS = sys.modules["builtins"]


def rsplit1(s, sep):
    """The same as s.rsplit(sep, 1), but works in 2.3"""
    parts = s.split(sep)
    return sep.join(parts[:-1]), parts[-1]


def run_python_module(modulename, args):
    """Run a python module, as though with ``python -m name args...``.

    `modulename` is the name of the module, possibly a dot-separated name.
    `args` is the argument array to present as sys.argv, including the first
    element naming the module being executed.

    """
    openfile = None
    glo, loc = globals(), locals()
    try:
        try:
            # Search for the module - inside its parent package, if any - using
            # standard import mechanics.
            if "." in modulename:
                packagename, name = rsplit1(modulename, ".")
                package = __import__(packagename, glo, loc, ["__path__"])
                searchpath = package.__path__
            else:
                packagename, name = None, modulename
                searchpath = None  # "top-level search" in imp.find_module()
            openfile, pathname, _ = imp.find_module(name, searchpath)

            # Complain if this is a magic non-file module.
            if openfile is None and pathname is None:
                raise NoSourceError("module does not live in a file: %r" % modulename)

            # If `modulename` is actually a package, not a mere module, then we
            # pretend to be Python 2.7 and try running its __main__.py script.
            if openfile is None:
                packagename = modulename
                name = "__main__"
                package = __import__(packagename, glo, loc, ["__path__"])
                searchpath = package.__path__
                openfile, pathname, _ = imp.find_module(name, searchpath)
        except ImportError:
            _, err, _ = sys.exc_info()
            raise NoSourceError(str(err))
    finally:
        if openfile:
            openfile.close()

    # Finally, hand the file off to run_python_file for execution.
    args[0] = pathname
    run_python_file(pathname, args, package=packagename)


def run_python_file(filename, args, package=None, callback=None):
    """Run a python file as if it were the main program on the command line.

    `filename` is the path to the file to execute, it need not be a .py file.
    `args` is the argument array to present as sys.argv, including the first
    element naming the file being executed.  `package` is the name of the
    enclosing package, if any.

    If `callback` is not None, it is a function which is called back as the
    execution progresses. This can be used for example in a debugger, or
    for custom tracing or statistics gathering.
    """
    # Create a module to serve as __main__
    old_main_mod = sys.modules["__main__"]
    main_mod = imp.new_module("__main__")
    sys.modules["__main__"] = main_mod
    main_mod.__file__ = filename
    if package:
        main_mod.__package__ = package
    main_mod.__builtins__ = BUILTINS

    # set sys.argv and the first path element properly.
    old_argv = sys.argv
    old_path0 = sys.path[0]

    # note: the type of args is na tuple; we want type(sys.argv) == list
    sys.argv = [filename] + list(args)

    if package:
        sys.path[0] = ""
    else:
        sys.path[0] = os.path.abspath(os.path.dirname(filename))

    is_pypy = IS_PYPY
    try:
        # Open the source or bytecode file.
        try:
            mime = mimetypes.guess_type(filename)
            if mime == ("application/x-python-code", None):
                (
                    python_version,
                    timestamp,
                    magic_int,
                    code,
                    is_pypy,
                    source_size,
                    sip_hash,
                ) = load_module(filename)
                supported_versions, mess = get_supported_versions(
                    is_pypy, is_bytecode=True
                )
                if python_version not in supported_versions:
                    raise WrongBytecodeError(
                        "We only support byte code for %s: %r is %2.1f bytecode"
                        % (mess, filename, python_version)
                    )
                pass
            else:
                source_file = open_source(filename)
                try:
                    source = source_file.read()
                finally:
                    source_file.close()

                supported_versions, mess = get_supported_versions(
                    IS_PYPY, is_bytecode=False
                )
                if PYTHON_VERSION not in supported_versions:
                    raise CannotCompileError(
                        "We need %s to compile source code; you are running Python %s"
                        % (mess, PYTHON_VERSION)
                    )

                # We have the source.  `compile` still needs the last line to be clean,
                # so make sure it is, then compile a code object from it.
                if not source or source[-1] != "\n":
                    source += "\n"
                code = compile(source, filename, "exec")
                python_version = PYTHON_VERSION

        except (IOError):
            raise NoSourceError("No file to run: %r" % filename)

        # Execute the source file.
        exec_code_object(code, main_mod.__dict__, python_version, is_pypy, callback)

    finally:
        # Restore the old __main__
        sys.modules["__main__"] = old_main_mod

        # Restore the old argv and path
        sys.argv = old_argv
        sys.path[0] = old_path0


def run_python_string(source, args, package=None, callback=None):
    """Run a python string as if it were the main program on the command line.
    """
    # Create a module to serve as __main__
    old_main_mod = sys.modules["__main__"]
    main_mod = imp.new_module("__main__")
    sys.modules["__main__"] = main_mod
    fake_path = main_mod.__file__ = "<string %s>" % source[:20]
    if package:
        main_mod.__package__ = package
    main_mod.__builtins__ = BUILTINS

    # Set sys.argv and the first path element properly.
    old_path0 = sys.path[0]
    sys.argv = [fake_path] + list(args)

    try:
        supported_versions, mess = get_supported_versions(IS_PYPY, is_bytecode=False)
        if PYTHON_VERSION not in supported_versions:
            raise CannotCompileError(
                "We need %s to compile source code; you are running %s"
                % (mess, PYTHON_VERSION)
            )

        # `compile` still needs the last line to be clean,
        # so make sure it is, then compile a code object from it.
        if not source or source[-1] != "\n":
            source += "\n"
        code = compile(source, fake_path, "exec")
        python_version = PYTHON_VERSION

        # Execute the source string.
        exec_code_object(code, main_mod.__dict__, python_version, IS_PYPY, callback)

    finally:
        # Restore the old __main__
        sys.modules["__main__"] = old_main_mod

        # Restore the old argv and path
        sys.path[0] = old_path0


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("usage: execfile.py <filename> args")
        sys.exit(1)
    run_python_file(sys.argv[1], sys.argv[2:])



"""Basic Python interpreter tests for x-python."""
from __future__ import print_function

try:
    import vmtest
except ImportError:
    from . import vmtest

from xdis import PYTHON3, PYTHON_VERSION
PY2 = not PYTHON3

class TestBasic(vmtest.VmTestCase):

    def test_attribute_access(self):
        self.self_checking()

        self.assert_ok("""\
            class Thing2(object):
                z = 17
                def __init__(self):
                    self.x = 23
            t = Thing2()
            print(t.xyzzy)
            """, raises=AttributeError)

    def test_bound_method_on_falsy_objects(self):
        self.self_checking()

    def test_comprehensions(self):
        self.self_checking()

    def test_inplace_operators(self):
        self.self_checking()

    def test_slice(self):
        self.self_checking()

    def test_slice_stmts(self):
        self.self_checking()

    def test_generator_expression(self):
        self.do_one()

    if PYTHON_VERSION in (3.6, 3.7):
        print("Test not gone over yet for %s" % PYTHON_VERSION)
    else:

        if PY2:
            def test_inplace_division(self):
                self.assert_ok("""\
                    x, y = 24, 3
                    x /= y
                    assert x == 8 and y == 3
                    assert isinstance(x, int)
                    x /= y
                    assert x == 2 and y == 3
                    assert isinstance(x, int)
                    """)
        elif PYTHON3:
            def test_inplace_division(self):
                self.assert_ok("""\
                    x, y = 24, 3
                    x /= y
                    assert x == 8.0 and y == 3
                    assert isinstance(x, float)
                    x /= y
                    assert x == (8.0/3.0) and y == 3
                    assert isinstance(x, float)
                    """)

        def test_building_stuff(self):
            self.assert_ok("""\
                print((1+1, 2+2, 3+3))
                """)
            self.assert_ok("""\
                print([1+1, 2+2, 3+3])
                """)
            self.assert_ok("""\
                print({1:1+1, 2:2+2, 3:3+3})
                """)

        def test_subscripting(self):
            self.assert_ok("""\
                l = list(range(10))
                print("%s %s %s" % (l[0], l[3], l[9]))
                """)
            self.assert_ok("""\
                l = list(range(10))
                l[5] = 17
                print(l)
                """)
            self.assert_ok("""\
                l = list(range(10))
                del l[5]
                print(l)
                """)

        def test_strange_sequence_ops(self):
            # from stdlib: test/test_augassign.py
            self.do_one()

        def test_unary_operators(self):
            self.assert_ok("""\
                x = 8
                print(-x, ~x, not x)
                """)

        def test_attribute_inplace_ops(self):
            self.assert_ok("""\
                l = lambda: 1   # Just to have an object...
                l.foo = 17
                l.foo -= 3
                print(l.foo)
                """)

        def test_deleting_names(self):
            self.assert_ok("""\
                g = 17
                assert g == 17
                del g
                g
                """, raises=NameError)

        def test_deleting_local_names(self):
            self.assert_ok("""\
                def f():
                    l = 23
                    assert l == 23
                    del l
                    l
                f()
                """, raises=NameError)

        def test_import(self):
            self.assert_ok("""\
                import math
                print(math.pi, math.e)
                from math import sqrt
                print(sqrt(2))
                from math import *
                print(sin(2))
                """)

        def test_classes(self):
            self.assert_ok("""\
                class Thing(object):
                    def __init__(self, x):
                        self.x = x
                    def meth(self, y):
                        return self.x * y
                thing1 = Thing(2)
                thing2 = Thing(3)
                print(thing1.x, thing2.x)
                print(thing1.meth(4), thing2.meth(5))
                """)

        def test_calling_methods_wrong(self):
            self.assert_ok("""\
                class Thing(object):
                    def __init__(self, x):
                        self.x = x
                    def meth(self, y):
                        return self.x * y
                thing1 = Thing(2)
                print(Thing.meth(14))
                """, raises=TypeError)

        def test_calling_subclass_methods(self):
            self.assert_ok("""\
                class Thing(object):
                    def foo(self):
                        return 17

                class SubThing(Thing):
                    pass

                st = SubThing()
                print(st.foo())
                """)

        def test_subclass_attribute(self):
            self.assert_ok("""\
                class Thing(object):
                    def __init__(self):
                        self.foo = 17
                class SubThing(Thing):
                    pass
                st = SubThing()
                print(st.foo)
                """)

        def test_subclass_attributes_not_shared(self):
            self.assert_ok("""\
                class Thing(object):
                    foo = 17
                class SubThing(Thing):
                    foo = 25
                st = SubThing()
                t = Thing()
                assert st.foo == 25
                assert t.foo == 17
                """)

        def test_object_attrs_not_shared_with_class(self):
            self.assert_ok("""\
                class Thing(object):
                    pass
                t = Thing()
                t.foo = 1
                Thing.foo""", raises=AttributeError)

        def test_data_descriptors_precede_instance_attributes(self):
            self.assert_ok("""\
                class Foo(object):
                    pass
                f = Foo()
                f.des = 3
                class Descr(object):
                    def __get__(self, obj, cls=None):
                        return 2
                    def __set__(self, obj, val):
                        raise NotImplementedError
                Foo.des = Descr()
                assert f.des == 2
                """)

        def test_instance_attrs_precede_non_data_descriptors(self):
            self.assert_ok("""\
                class Foo(object):
                    pass
                f = Foo()
                f.des = 3
                class Descr(object):
                    def __get__(self, obj, cls=None):
                        return 2
                Foo.des = Descr()
                assert f.des == 3
                """)

        def test_subclass_attributes_dynamic(self):
            self.assert_ok("""\
                class Foo(object):
                    pass
                class Bar(Foo):
                    pass
                b = Bar()
                Foo.baz = 3
                assert b.baz == 3
                """)

        def test_staticmethods(self):
            self.assert_ok("""\
                class Thing(object):
                    @staticmethod
                    def smeth(x):
                        print(x)
                    @classmethod
                    def cmeth(cls, x):
                        print(x)

                Thing.smeth(1492)
                Thing.cmeth(1776)
                """)

        def test_unbound_methods(self):
            self.assert_ok("""\
                class Thing(object):
                    def meth(self, x):
                        print(x)
                m = Thing.meth
                m(Thing(), 1815)
                """)

        def test_bound_methods(self):
            self.assert_ok("""\
                class Thing(object):
                    def meth(self, x):
                        print(x)
                t = Thing()
                m = t.meth
                m(1815)
                """)

        def test_callback(self):
            self.assert_ok("""\
                def lcase(s):
                    return s.lower()
                l = ["xyz", "ABC"]
                l.sort(key=lcase)
                print(l)
                assert l == ["ABC", "xyz"]
                """)

        def test_unpacking(self):
            self.assert_ok("""\
                a, b, c = (1, 2, 3)
                assert a == 1
                assert b == 2
                assert c == 3
                """)

        if PY2:
            def test_exec_statement(self):
                self.assert_ok("""\
                    g = {}
                    exec "a = 11" in g, g
                    assert g['a'] == 11
                    """)
        elif PYTHON3:
            def test_exec_statement(self):
                self.assert_ok("""\
                    g = {}
                    exec("a = 11", g, g)
                    assert g['a'] == 11
                    """)

        def test_jump_if_true_or_pop(self):
            self.assert_ok("""\
                def f(a, b):
                    return a or b
                assert f(17, 0) == 17
                assert f(0, 23) == 23
                assert f(0, "") == ""
                """)

        def test_jump_if_false_or_pop(self):
            self.assert_ok("""\
                def f(a, b):
                    return not(a and b)
                assert f(17, 0) is True
                assert f(0, 23) is True
                assert f(0, "") is True
                assert f(17, 23) is False
                """)

        def test_pop_jump_if_true(self):
            self.assert_ok("""\
                def f(a):
                    if not a:
                        return 'foo'
                    else:
                        return 'bar'
                assert f(0) == 'foo'
                assert f(1) == 'bar'
                """)

        def test_multiple_classes(self):
            # Making classes used to mix together all the class-scoped values
            # across classes.  This test would fail because A.__init__ would be
            # over-written with B.__init__, and A(1, 2, 3) would complain about
            # too many arguments.
            self.assert_ok("""\
                class A(object):
                    def __init__(self, a, b, c):
                        self.sum = a + b + c

                class B(object):
                    def __init__(self, x):
                        self.x = x

                a = A(1, 2, 3)
                b = B(7)
                print(a.sum)
                print(b.x)
                """)


    if PY2:
        class TestPrinting(vmtest.VmTestCase):
            def test_printing(self):
                self.assert_ok("print 'hello'")
                self.assert_ok("a = 3; print a+4")
                self.assert_ok("""
                    print 'hi', 17, u'bye', 23,
                    print "", "\t", "the end"
                    """)

            def test_printing_in_a_function(self):
                self.assert_ok("""\
                    def fn():
                        print "hello"
                    fn()
                    print "bye"
                    """)

            def test_printing_to_a_file(self):
                self.assert_ok("""\
                    import sys
                    print >>sys.stdout, 'hello', 'there'
                    """)


    class TestLoops(vmtest.VmTestCase):
        def test_break(self):
            self.assert_ok("""\
                for i in range(10):
                    print(i)
                    if i == 7:
                        break
                print("done")
                """)

        def test_continue(self):
            # fun fact: this doesn't use CONTINUE_LOOP
            self.assert_ok("""\
                for i in range(10):
                    if i % 3 == 0:
                        continue
                    print(i)
                print("done")
                """)

        def test_continue_in_try_except(self):
            self.assert_ok("""\
                for i in range(10):
                    try:
                        if i % 3 == 0:
                            continue
                        print(i)
                    except ValueError:
                        pass
                print("done")
                """)

        def test_continue_in_try_finally(self):
            self.assert_ok("""\
                for i in range(10):
                    try:
                        if i % 3 == 0:
                            continue
                        print(i)
                    finally:
                        print(".")
                print("done")
                """)


    class TestComparisons(vmtest.VmTestCase):
        def test_in(self):
            self.assert_ok("""\
                assert "x" in "xyz"
                assert "x" not in "abc"
                assert "x" in ("x", "y", "z")
                assert "x" not in ("a", "b", "c")
                """)

        def test_less(self):
            self.assert_ok("""\
                assert 1 < 3
                assert 1 <= 2 and 1 <= 1
                assert "a" < "b"
                assert "a" <= "b" and "a" <= "a"
                """)

        def test_greater(self):
            self.assert_ok("""\
                assert 3 > 1
                assert 3 >= 1 and 3 >= 3
                assert "z" > "a"
                assert "z" >= "a" and "z" >= "z"
                """)

if __name__ == "__main__":
    # import unittest
    # unittest.main()

    t = TestBasic("test_decorator")
    # t.test_decorator()
    # t = TestComparisons("test_in")
    # t.test_in()
    # t = TestComparisons("test_greater")
    # t.test_greater()



# In addition to testing attributes, this
# is a simple class test.

class Thing(object):
    z = 17
    def __init__(self):
        self.x = 23
t = Thing()

assert Thing.z == 17
assert t.z == 17
assert t.x == 23
